#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_user_federation

short_description: Allows administration of Keycloak user federations using Keycloak API

version_added: 3.7.0

description:
  - This module allows you to add, remove or modify Keycloak user federations using the Keycloak REST API. It requires access
    to the REST API using OpenID Connect; the user connecting and the client being used must have the requisite access rights.
    In a default Keycloak installation, admin-cli and an admin user would work, as would a separate client definition with
    the scope tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/20.0.2/rest-api/index.html).
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the user federation.
      - On V(present), the user federation is created if it does not yet exist, or updated with the parameters you provide.
      - On V(absent), the user federation is removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  realm:
    description:
      - The Keycloak realm under which this user federation resides.
    default: 'master'
    type: str

  id:
    description:
      - The unique ID for this user federation. If left empty, the user federation is searched by its O(name).
    type: str

  name:
    description:
      - Display name of provider when linked in admin console.
    type: str

  provider_id:
    description:
      - Provider for this user federation. Built-in providers are V(ldap), V(kerberos), and V(sssd). Custom user storage providers
        can also be used.
    aliases:
      - providerId
    type: str

  provider_type:
    description:
      - Component type for user federation (only supported value is V(org.keycloak.storage.UserStorageProvider)).
    aliases:
      - providerType
    default: org.keycloak.storage.UserStorageProvider
    type: str

  parent_id:
    description:
      - Unique ID for the parent of this user federation. Realm ID is automatically used if left blank.
    aliases:
      - parentId
    type: str

  remove_unspecified_mappers:
    description:
      - Remove mappers that are not specified in the configuration for this federation.
      - Set to V(false) to keep mappers that are not listed in O(mappers).
    type: bool
    default: true
    version_added: 9.4.0

  bind_credential_update_mode:
    description:
      - The value of the config parameter O(config.bindCredential) is redacted in the Keycloak responses. Comparing the redacted
        value with the desired value always evaluates to not equal. This means the before and desired states are never equal
        if the parameter is set.
      - Set to V(always) to include O(config.bindCredential) in the comparison of before and desired state. Because of the
        redacted value returned by Keycloak the module always detects a change and make an update if a O(config.bindCredential)
        value is set.
      - Set to V(only_indirect) to exclude O(config.bindCredential) when comparing the before state with the desired state.
        The value of O(config.bindCredential) is only updated if there are other changes to the user federation that require
        an update.
    type: str
    default: always
    choices:
      - always
      - only_indirect
    version_added: 9.5.0

  config:
    description:
      - Dict specifying the configuration options for the provider; the contents differ depending on the value of O(provider_id).
        Examples are given below for V(ldap), V(kerberos) and V(sssd). It is easiest to obtain valid config values by dumping
        an already-existing user federation configuration through check-mode in the RV(existing) field.
      - The value V(sssd) has been supported since community.general 4.2.0.
    type: dict
    suboptions:
      enabled:
        description:
          - Enable/disable this user federation.
        default: true
        type: bool

      priority:
        description:
          - Priority of provider when doing a user lookup. Lowest first.
        default: 0
        type: int

      importEnabled:
        description:
          - If V(true), LDAP users are imported into Keycloak DB and synced by the configured sync policies.
        default: true
        type: bool

      editMode:
        description:
          - V(READ_ONLY) is a read-only LDAP store. V(WRITABLE) means data is synced back to LDAP on demand. V(UNSYNCED) means
            user data is imported, but not synced back to LDAP.
        type: str
        choices:
          - READ_ONLY
          - WRITABLE
          - UNSYNCED

      syncRegistrations:
        description:
          - Should newly created users be created within LDAP store? Priority effects which provider is chosen to sync the
            new user.
        default: false
        type: bool

      vendor:
        description:
          - LDAP vendor (provider).
          - Use short name. For instance, write V(rhds) for "Red Hat Directory Server".
        type: str

      usernameLDAPAttribute:
        description:
          - Name of LDAP attribute, which is mapped as Keycloak username. For many LDAP server vendors it can be V(uid). For
            Active directory it can be V(sAMAccountName) or V(cn). The attribute should be filled for all LDAP user records
            you want to import from LDAP to Keycloak.
        type: str

      rdnLDAPAttribute:
        description:
          - Name of LDAP attribute, which is used as RDN (top attribute) of typical user DN. Usually it is the same as Username
            LDAP attribute, however it is not required. For example for Active directory, it is common to use V(cn) as RDN
            attribute when username attribute might be V(sAMAccountName).
        type: str

      uuidLDAPAttribute:
        description:
          - Name of LDAP attribute, which is used as unique object identifier (UUID) for objects in LDAP. For many LDAP server
            vendors, it is V(entryUUID); however some are different. For example for Active directory it should be V(objectGUID).
            If your LDAP server does not support the notion of UUID, you can use any other attribute that is supposed to be
            unique among LDAP users in tree.
        type: str

      userObjectClasses:
        description:
          - All values of LDAP objectClass attribute for users in LDAP divided by comma. For example V(inetOrgPerson, organizationalPerson).
            Newly created Keycloak users are written to LDAP with all those object classes and existing LDAP user records
            are found just if they contain all those object classes.
        type: str

      connectionUrl:
        description:
          - Connection URL to your LDAP server.
        type: str

      usersDn:
        description:
          - Full DN of LDAP tree where your users are. This DN is the parent of LDAP users.
        type: str

      customUserSearchFilter:
        description:
          - Additional LDAP Filter for filtering searched users. Leave this empty if you do not need additional filter.
        type: str

      searchScope:
        description:
          - For one level, the search applies only for users in the DNs specified by User DNs. For subtree, the search applies
            to the whole subtree. See LDAP documentation for more details.
        default: '1'
        type: str
        choices:
          - '1'
          - '2'

      authType:
        description:
          - Type of the Authentication method used during LDAP Bind operation. It is used in most of the requests sent to
            the LDAP server.
        default: 'none'
        type: str
        choices:
          - none
          - simple

      bindDn:
        description:
          - DN of LDAP user which is used by Keycloak to access LDAP server.
        type: str

      bindCredential:
        description:
          - Password of LDAP admin.
        type: str

      startTls:
        description:
          - Encrypts the connection to LDAP using STARTTLS, which disables connection pooling.
        default: false
        type: bool

      usePasswordModifyExtendedOp:
        description:
          - Use the LDAPv3 Password Modify Extended Operation (RFC-3062). The password modify extended operation usually requires
            that LDAP user already has password in the LDAP server. So when this is used with 'Sync Registrations', it can
            be good to add also 'Hardcoded LDAP attribute mapper' with randomly generated initial password.
        default: false
        type: bool

      validatePasswordPolicy:
        description:
          - Determines if Keycloak should validate the password with the realm password policy before updating it.
        default: false
        type: bool

      trustEmail:
        description:
          - If enabled, email provided by this provider is not verified even if verification is enabled for the realm.
        default: false
        type: bool

      useTruststoreSpi:
        description:
          - Specifies whether LDAP connection uses the truststore SPI with the truststore configured in standalone.xml/domain.xml.
            V(always) means that it always uses it. V(never) means that it does not use it. V(ldapsOnly) means that it uses
            if your connection URL use ldaps.
          - Note even if standalone.xml/domain.xml is not configured, the default Java cacerts or certificate specified by
            C(javax.net.ssl.trustStore) property is used.
        default: ldapsOnly
        type: str
        choices:
          - always
          - ldapsOnly
          - never

      connectionTimeout:
        description:
          - LDAP Connection Timeout in milliseconds.
        type: int

      readTimeout:
        description:
          - LDAP Read Timeout in milliseconds. This timeout applies for LDAP read operations.
        type: int

      pagination:
        description:
          - Does the LDAP server support pagination.
        default: true
        type: bool

      connectionPooling:
        description:
          - Determines if Keycloak should use connection pooling for accessing LDAP server.
        default: true
        type: bool

      connectionPoolingAuthentication:
        description:
          - A list of space-separated authentication types of connections that may be pooled.
        type: str
        choices:
          - none
          - simple
          - DIGEST-MD5

      connectionPoolingDebug:
        description:
          - A string that indicates the level of debug output to produce. Example valid values are V(fine) (trace connection
            creation and removal) and V(all) (all debugging information).
        type: str

      connectionPoolingInitSize:
        description:
          - The number of connections per connection identity to create when initially creating a connection for the identity.
        type: int

      connectionPoolingMaxSize:
        description:
          - The maximum number of connections per connection identity that can be maintained concurrently.
        type: int

      connectionPoolingPrefSize:
        description:
          - The preferred number of connections per connection identity that should be maintained concurrently.
        type: int

      connectionPoolingProtocol:
        description:
          - A list of space-separated protocol types of connections that may be pooled. Valid types are V(plain) and V(ssl).
        type: str

      connectionPoolingTimeout:
        description:
          - The number of milliseconds that an idle connection may remain in the pool without being closed and removed from
            the pool.
        type: int

      allowKerberosAuthentication:
        description:
          - Enable/disable HTTP authentication of users with SPNEGO/Kerberos tokens. The data about authenticated users is
            provisioned from this LDAP server.
        default: false
        type: bool

      kerberosRealm:
        description:
          - Name of kerberos realm.
        type: str

      krbPrincipalAttribute:
        description:
          - Name of the LDAP attribute, which refers to Kerberos principal. This is used to lookup appropriate LDAP user after
            successful Kerberos/SPNEGO authentication in Keycloak. When this is empty, the LDAP user is looked up based on
            LDAP username corresponding to the first part of his Kerberos principal. For instance, for principal C(john@KEYCLOAK.ORG),
            it assumes that LDAP username is V(john).
        type: str
        version_added: 8.1.0

      serverPrincipal:
        description:
          - Full name of server principal for HTTP service including server and domain name. For example V(HTTP/host.foo.org@FOO.ORG).
            Use V(*) to accept any service principal in the KeyTab file.
        type: str

      keyTab:
        description:
          - Location of Kerberos KeyTab file containing the credentials of server principal. For example V(/etc/krb5.keytab).
        type: str

      debug:
        description:
          - Enable/disable debug logging to standard output for Krb5LoginModule.
        type: bool

      useKerberosForPasswordAuthentication:
        description:
          - Use Kerberos login module for authenticate username/password against Kerberos server instead of authenticating
            against LDAP server with Directory Service API.
        default: false
        type: bool

      allowPasswordAuthentication:
        description:
          - Enable/disable possibility of username/password authentication against Kerberos database.
        type: bool

      batchSizeForSync:
        description:
          - Count of LDAP users to be imported from LDAP to Keycloak within a single transaction.
        default: 1000
        type: int

      fullSyncPeriod:
        description:
          - Period for full synchronization in seconds.
        default: -1
        type: int

      changedSyncPeriod:
        description:
          - Period for synchronization of changed or newly created LDAP users in seconds.
        default: -1
        type: int

      updateProfileFirstLogin:
        description:
          - Update profile on first login.
        type: bool

      cachePolicy:
        description:
          - Cache Policy for this storage provider.
        type: str
        default: 'DEFAULT'
        choices:
          - DEFAULT
          - EVICT_DAILY
          - EVICT_WEEKLY
          - MAX_LIFESPAN
          - NO_CACHE

      evictionDay:
        description:
          - Day of the week the entry is set to become invalid on.
        type: str

      evictionHour:
        description:
          - Hour of day the entry is set to become invalid on.
        type: str

      evictionMinute:
        description:
          - Minute of day the entry is set to become invalid on.
        type: str

      maxLifespan:
        description:
          - Max lifespan of cache entry in milliseconds.
        type: int

      referral:
        description:
          - Specifies if LDAP referrals should be followed or ignored. Please note that enabling referrals can slow down authentication
            as it allows the LDAP server to decide which other LDAP servers to use. This could potentially include untrusted
            servers.
        type: str
        choices:
          - ignore
          - follow
        version_added: 9.5.0

  mappers:
    description:
      - A list of dicts defining mappers associated with this Identity Provider.
    type: list
    elements: dict
    suboptions:
      id:
        description:
          - Unique ID of this mapper.
        type: str

      name:
        description:
          - Name of the mapper. If no ID is given, the mapper is searched by name.
        type: str

      parentId:
        description:
          - Unique ID for the parent of this mapper. ID of the user federation is automatically used if left blank.
        type: str

      providerId:
        description:
          - The mapper type for this mapper (for instance V(user-attribute-ldap-mapper)).
        type: str

      providerType:
        description:
          - Component type for this mapper.
        type: str
        default: org.keycloak.storage.ldap.mappers.LDAPStorageMapper

      config:
        description:
          - Dict specifying the configuration options for the mapper; the contents differ depending on the value of I(identityProviderMapper).
        type: dict

extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Laurent Paumier (@laurpaum)
"""

EXAMPLES = r"""
- name: Create LDAP user federation
  community.general.keycloak_user_federation:
    auth_keycloak_url: https://keycloak.example.com/auth
    auth_realm: master
    auth_username: admin
    auth_password: password
    realm: my-realm
    name: my-ldap
    state: present
    provider_id: ldap
    provider_type: org.keycloak.storage.UserStorageProvider
    config:
      priority: 0
      enabled: true
      cachePolicy: DEFAULT
      batchSizeForSync: 1000
      editMode: READ_ONLY
      importEnabled: true
      syncRegistrations: false
      vendor: other
      usernameLDAPAttribute: uid
      rdnLDAPAttribute: uid
      uuidLDAPAttribute: entryUUID
      userObjectClasses: inetOrgPerson, organizationalPerson
      connectionUrl: ldaps://ldap.example.com:636
      usersDn: ou=Users,dc=example,dc=com
      authType: simple
      bindDn: cn=directory reader
      bindCredential: password
      searchScope: 1
      validatePasswordPolicy: false
      trustEmail: false
      useTruststoreSpi: ldapsOnly
      connectionPooling: true
      pagination: true
      allowKerberosAuthentication: false
      debug: false
      useKerberosForPasswordAuthentication: false
    mappers:
      - name: "full name"
        providerId: "full-name-ldap-mapper"
        providerType: "org.keycloak.storage.ldap.mappers.LDAPStorageMapper"
        config:
          ldap.full.name.attribute: cn
          read.only: true
          write.only: false

- name: Create Kerberos user federation
  community.general.keycloak_user_federation:
    auth_keycloak_url: https://keycloak.example.com/auth
    auth_realm: master
    auth_username: admin
    auth_password: password
    realm: my-realm
    name: my-kerberos
    state: present
    provider_id: kerberos
    provider_type: org.keycloak.storage.UserStorageProvider
    config:
      priority: 0
      enabled: true
      cachePolicy: DEFAULT
      kerberosRealm: EXAMPLE.COM
      serverPrincipal: HTTP/host.example.com@EXAMPLE.COM
      keyTab: keytab
      allowPasswordAuthentication: false
      updateProfileFirstLogin: false

- name: Create sssd user federation
  community.general.keycloak_user_federation:
    auth_keycloak_url: https://keycloak.example.com/auth
    auth_realm: master
    auth_username: admin
    auth_password: password
    realm: my-realm
    name: my-sssd
    state: present
    provider_id: sssd
    provider_type: org.keycloak.storage.UserStorageProvider
    config:
      priority: 0
      enabled: true
      cachePolicy: DEFAULT

- name: Delete user federation
  community.general.keycloak_user_federation:
    auth_keycloak_url: https://keycloak.example.com/auth
    auth_realm: master
    auth_username: admin
    auth_password: password
    realm: my-realm
    name: my-federation
    state: absent
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "No changes required to user federation 164bb483-c613-482e-80fe-7f1431308799."

proposed:
  description: Representation of proposed user federation.
  returned: always
  type: dict
  sample:
    {
      "config": {
        "allowKerberosAuthentication": "false",
        "authType": "simple",
        "batchSizeForSync": "1000",
        "bindCredential": "**********",
        "bindDn": "cn=directory reader",
        "cachePolicy": "DEFAULT",
        "connectionPooling": "true",
        "connectionUrl": "ldaps://ldap.example.com:636",
        "debug": "false",
        "editMode": "READ_ONLY",
        "enabled": "true",
        "importEnabled": "true",
        "pagination": "true",
        "priority": "0",
        "rdnLDAPAttribute": "uid",
        "searchScope": "1",
        "syncRegistrations": "false",
        "trustEmail": "false",
        "useKerberosForPasswordAuthentication": "false",
        "useTruststoreSpi": "ldapsOnly",
        "userObjectClasses": "inetOrgPerson, organizationalPerson",
        "usernameLDAPAttribute": "uid",
        "usersDn": "ou=Users,dc=example,dc=com",
        "uuidLDAPAttribute": "entryUUID",
        "validatePasswordPolicy": "false",
        "vendor": "other"
      },
      "name": "ldap",
      "providerId": "ldap",
      "providerType": "org.keycloak.storage.UserStorageProvider"
    }

existing:
  description: Representation of existing user federation.
  returned: always
  type: dict
  sample:
    {
      "config": {
        "allowKerberosAuthentication": "false",
        "authType": "simple",
        "batchSizeForSync": "1000",
        "bindCredential": "**********",
        "bindDn": "cn=directory reader",
        "cachePolicy": "DEFAULT",
        "changedSyncPeriod": "-1",
        "connectionPooling": "true",
        "connectionUrl": "ldaps://ldap.example.com:636",
        "debug": "false",
        "editMode": "READ_ONLY",
        "enabled": "true",
        "fullSyncPeriod": "-1",
        "importEnabled": "true",
        "pagination": "true",
        "priority": "0",
        "rdnLDAPAttribute": "uid",
        "searchScope": "1",
        "syncRegistrations": "false",
        "trustEmail": "false",
        "useKerberosForPasswordAuthentication": "false",
        "useTruststoreSpi": "ldapsOnly",
        "userObjectClasses": "inetOrgPerson, organizationalPerson",
        "usernameLDAPAttribute": "uid",
        "usersDn": "ou=Users,dc=example,dc=com",
        "uuidLDAPAttribute": "entryUUID",
        "validatePasswordPolicy": "false",
        "vendor": "other"
      },
      "id": "01122837-9047-4ae4-8ca0-6e2e891a765f",
      "mappers": [
        {
          "config": {
            "always.read.value.from.ldap": "false",
            "is.mandatory.in.ldap": "false",
            "ldap.attribute": "mail",
            "read.only": "true",
            "user.model.attribute": "email"
          },
          "id": "17d60ce2-2d44-4c2c-8b1f-1fba601b9a9f",
          "name": "email",
          "parentId": "01122837-9047-4ae4-8ca0-6e2e891a765f",
          "providerId": "user-attribute-ldap-mapper",
          "providerType": "org.keycloak.storage.ldap.mappers.LDAPStorageMapper"
        }
      ],
      "name": "myfed",
      "parentId": "myrealm",
      "providerId": "ldap",
      "providerType": "org.keycloak.storage.UserStorageProvider"
    }

end_state:
  description: Representation of user federation after module execution.
  returned: on success
  type: dict
  sample:
    {
      "config": {
        "allowPasswordAuthentication": "false",
        "cachePolicy": "DEFAULT",
        "enabled": "true",
        "kerberosRealm": "EXAMPLE.COM",
        "keyTab": "/etc/krb5.keytab",
        "priority": "0",
        "serverPrincipal": "HTTP/host.example.com@EXAMPLE.COM",
        "updateProfileFirstLogin": "false"
      },
      "id": "cf52ae4f-4471-4435-a0cf-bb620cadc122",
      "mappers": [],
      "name": "kerberos",
      "parentId": "myrealm",
      "providerId": "kerberos",
      "providerType": "org.keycloak.storage.UserStorageProvider"
    }
"""

from copy import deepcopy
from urllib.parse import urlencode

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    camel,
    get_token,
    keycloak_argument_spec,
)


def normalize_kc_comp(comp):
    if "config" in comp:
        # kc completely removes the parameter `krbPrincipalAttribute` if it is set to `''`; the unset kc parameter is equivalent to `''`;
        # to make change detection and diff more accurate we set it again in the kc responses
        if "krbPrincipalAttribute" not in comp["config"]:
            comp["config"]["krbPrincipalAttribute"] = [""]

        # kc stores a timestamp of the last sync in `lastSync` to time the periodic sync, it is removed to minimize diff/changes
        comp["config"].pop("lastSync", None)


def sanitize(comp):
    def sanitize_value(v):
        """Convert list values: single-element lists to strings, multi-element lists sorted alphabetically, others as-is."""
        if isinstance(v, list):
            if len(v) == 0:
                return None
            elif len(v) == 1:
                return v[0]
            else:
                return sorted(v)
        else:
            return v

    compcopy = deepcopy(comp)
    if "config" in compcopy:
        compcopy["config"] = {k: sanitize_value(v) for k, v in compcopy["config"].items()}
        # Remove None values (empty lists converted)
        compcopy["config"] = {k: v for k, v in compcopy["config"].items() if v is not None}
        if "bindCredential" in compcopy["config"]:
            compcopy["config"]["bindCredential"] = "**********"
    if "mappers" in compcopy:
        for mapper in compcopy["mappers"]:
            if "config" in mapper:
                mapper["config"] = {k: sanitize_value(v) for k, v in mapper["config"].items()}
                mapper["config"] = {k: v for k, v in mapper["config"].items() if v is not None}
    return compcopy


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    config_spec = dict(
        allowKerberosAuthentication=dict(type="bool", default=False),
        allowPasswordAuthentication=dict(type="bool"),
        authType=dict(type="str", choices=["none", "simple"], default="none"),
        batchSizeForSync=dict(type="int", default=1000),
        bindCredential=dict(type="str", no_log=True),
        bindDn=dict(type="str"),
        cachePolicy=dict(
            type="str",
            choices=["DEFAULT", "EVICT_DAILY", "EVICT_WEEKLY", "MAX_LIFESPAN", "NO_CACHE"],
            default="DEFAULT",
        ),
        changedSyncPeriod=dict(type="int", default=-1),
        connectionPooling=dict(type="bool", default=True),
        connectionPoolingAuthentication=dict(type="str", choices=["none", "simple", "DIGEST-MD5"]),
        connectionPoolingDebug=dict(type="str"),
        connectionPoolingInitSize=dict(type="int"),
        connectionPoolingMaxSize=dict(type="int"),
        connectionPoolingPrefSize=dict(type="int"),
        connectionPoolingProtocol=dict(type="str"),
        connectionPoolingTimeout=dict(type="int"),
        connectionTimeout=dict(type="int"),
        connectionUrl=dict(type="str"),
        customUserSearchFilter=dict(type="str"),
        debug=dict(type="bool"),
        editMode=dict(type="str", choices=["READ_ONLY", "WRITABLE", "UNSYNCED"]),
        enabled=dict(type="bool", default=True),
        evictionDay=dict(type="str"),
        evictionHour=dict(type="str"),
        evictionMinute=dict(type="str"),
        fullSyncPeriod=dict(type="int", default=-1),
        importEnabled=dict(type="bool", default=True),
        kerberosRealm=dict(type="str"),
        keyTab=dict(type="str", no_log=False),
        maxLifespan=dict(type="int"),
        pagination=dict(type="bool", default=True),
        priority=dict(type="int", default=0),
        rdnLDAPAttribute=dict(type="str"),
        readTimeout=dict(type="int"),
        referral=dict(type="str", choices=["ignore", "follow"]),
        searchScope=dict(type="str", choices=["1", "2"], default="1"),
        serverPrincipal=dict(type="str"),
        krbPrincipalAttribute=dict(type="str"),
        startTls=dict(type="bool", default=False),
        syncRegistrations=dict(type="bool", default=False),
        trustEmail=dict(type="bool", default=False),
        updateProfileFirstLogin=dict(type="bool"),
        useKerberosForPasswordAuthentication=dict(type="bool", default=False),
        usePasswordModifyExtendedOp=dict(type="bool", default=False, no_log=False),
        useTruststoreSpi=dict(type="str", choices=["always", "ldapsOnly", "never"], default="ldapsOnly"),
        userObjectClasses=dict(type="str"),
        usernameLDAPAttribute=dict(type="str"),
        usersDn=dict(type="str"),
        uuidLDAPAttribute=dict(type="str"),
        validatePasswordPolicy=dict(type="bool", default=False),
        vendor=dict(type="str"),
    )

    mapper_spec = dict(
        id=dict(type="str"),
        name=dict(type="str"),
        parentId=dict(type="str"),
        providerId=dict(type="str"),
        providerType=dict(type="str", default="org.keycloak.storage.ldap.mappers.LDAPStorageMapper"),
        config=dict(type="dict"),
    )

    meta_args = dict(
        config=dict(type="dict", options=config_spec),
        state=dict(type="str", default="present", choices=["present", "absent"]),
        realm=dict(type="str", default="master"),
        id=dict(type="str"),
        name=dict(type="str"),
        provider_id=dict(type="str", aliases=["providerId"]),
        provider_type=dict(type="str", aliases=["providerType"], default="org.keycloak.storage.UserStorageProvider"),
        parent_id=dict(type="str", aliases=["parentId"]),
        remove_unspecified_mappers=dict(type="bool", default=True),
        bind_credential_update_mode=dict(type="str", default="always", choices=["always", "only_indirect"]),
        mappers=dict(type="list", elements="dict", options=mapper_spec),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [
                ["id", "name"],
                ["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"],
            ]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", diff={}, proposed={}, existing={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    state = module.params.get("state")
    config = module.params.get("config")
    mappers = module.params.get("mappers")
    cid = module.params.get("id")
    name = module.params.get("name")

    # Keycloak API expects config parameters to be arrays containing a single string element
    if config is not None:
        module.params["config"] = {
            k: [str(v).lower() if not isinstance(v, str) else v] for k, v in config.items() if config[k] is not None
        }

    if mappers is not None:
        for mapper in mappers:
            if mapper.get("config") is not None:
                new_config = {}
                for k, v in mapper["config"].items():
                    if v is None:
                        continue
                    if isinstance(v, list):
                        new_config[k] = [str(item).lower() if not isinstance(item, str) else item for item in v]
                    else:
                        new_config[k] = [str(v).lower() if not isinstance(v, str) else v]
                mapper["config"] = new_config

    # Filter and map the parameters names that apply
    comp_params = [
        x
        for x in module.params
        if x
        not in list(keycloak_argument_spec().keys())
        + ["state", "realm", "mappers", "remove_unspecified_mappers", "bind_credential_update_mode"]
        and module.params.get(x) is not None
    ]

    # See if it already exists in Keycloak
    if cid is None:
        found = kc.get_components(urlencode(dict(type="org.keycloak.storage.UserStorageProvider", name=name)), realm)
        if len(found) > 1:
            module.fail_json(
                msg=f"No ID given and found multiple user federations with name `{name}`. Cannot continue."
            )
        before_comp = next(iter(found), None)
        if before_comp is not None:
            cid = before_comp["id"]
    else:
        before_comp = kc.get_component(cid, realm)

    if before_comp is None:
        before_comp = {}

    # if user federation exists, get associated mappers
    if cid is not None and before_comp:
        before_comp["mappers"] = sorted(
            kc.get_components(urlencode(dict(parent=cid)), realm), key=lambda x: x.get("name") or ""
        )

    normalize_kc_comp(before_comp)

    # Build a proposed changeset from parameters given to this module
    changeset = {}

    for param in comp_params:
        new_param_value = module.params.get(param)
        old_value = before_comp[camel(param)] if camel(param) in before_comp else None
        if param == "mappers":
            new_param_value = [{k: v for k, v in x.items() if v is not None} for x in new_param_value]
        if new_param_value != old_value:
            changeset[camel(param)] = new_param_value

    # special handling of mappers list to allow change detection
    if module.params.get("mappers") is not None:
        if module.params["provider_id"] in ["kerberos", "sssd"]:
            module.fail_json(msg=f"Cannot configure mappers for {module.params['provider_id']} provider.")
        for change in module.params["mappers"]:
            change = {k: v for k, v in change.items() if v is not None}
            if change.get("id") is None and change.get("name") is None:
                module.fail_json(msg="Either `name` or `id` has to be specified on each mapper.")
            if cid is None:
                old_mapper = {}
            elif change.get("id") is not None:
                old_mapper = next(
                    (
                        before_mapper
                        for before_mapper in before_comp.get("mappers", [])
                        if before_mapper["id"] == change["id"]
                    ),
                    None,
                )
                if old_mapper is None:
                    old_mapper = {}
            else:
                found = [
                    before_mapper
                    for before_mapper in before_comp.get("mappers", [])
                    if before_mapper["name"] == change["name"]
                ]
                if len(found) > 1:
                    module.fail_json(msg=f"Found multiple mappers with name `{change['name']}`. Cannot continue.")
                if len(found) == 1:
                    old_mapper = found[0]
                else:
                    old_mapper = {}
            new_mapper = old_mapper.copy()
            new_mapper.update(change)
            # changeset contains all desired mappers: those existing, to update or to create
            if changeset.get("mappers") is None:
                changeset["mappers"] = list()
            changeset["mappers"].append(new_mapper)
        changeset["mappers"] = sorted(changeset["mappers"], key=lambda x: x.get("name") or "")

        # to keep unspecified existing mappers we add them to the desired mappers list, unless they're already present
        if not module.params["remove_unspecified_mappers"] and "mappers" in before_comp:
            changeset_mapper_ids = [mapper["id"] for mapper in changeset["mappers"] if "id" in mapper]
            changeset["mappers"].extend(
                [mapper for mapper in before_comp["mappers"] if mapper["id"] not in changeset_mapper_ids]
            )

    # Prepare the desired values using the existing values (non-existence results in a dict that is save to use as a basis)
    desired_comp = before_comp.copy()
    desired_comp.update(changeset)

    result["proposed"] = sanitize(changeset)
    result["existing"] = sanitize(before_comp)

    # Cater for when it doesn't exist (an empty dict)
    if not before_comp:
        if state == "absent":
            # Do nothing and exit
            if module._diff:
                result["diff"] = dict(before="", after="")
            result["changed"] = False
            result["end_state"] = {}
            result["msg"] = "User federation does not exist; doing nothing."
            module.exit_json(**result)

        # Process a creation
        result["changed"] = True

        if module.check_mode:
            if module._diff:
                result["diff"] = dict(before="", after=sanitize(desired_comp))
            module.exit_json(**result)

        # create it
        desired_mappers = desired_comp.pop("mappers", [])
        after_comp = kc.create_component(desired_comp, realm)
        cid = after_comp["id"]
        updated_mappers = []
        # when creating a user federation, keycloak automatically creates default mappers
        default_mappers = kc.get_components(urlencode(dict(parent=cid)), realm)

        # create new mappers or update existing default mappers
        for desired_mapper in desired_mappers:
            found = [
                default_mapper for default_mapper in default_mappers if default_mapper["name"] == desired_mapper["name"]
            ]
            if len(found) > 1:
                module.fail_json(msg=f"Found multiple mappers with name `{desired_mapper['name']}`. Cannot continue.")
            if len(found) == 1:
                old_mapper = found[0]
            else:
                old_mapper = {}

            new_mapper = old_mapper.copy()
            new_mapper.update(desired_mapper)

            if new_mapper.get("id") is not None:
                kc.update_component(new_mapper, realm)
                updated_mappers.append(new_mapper)
            else:
                if new_mapper.get("parentId") is None:
                    new_mapper["parentId"] = cid
                updated_mappers.append(kc.create_component(new_mapper, realm))

        if module.params["remove_unspecified_mappers"]:
            # we remove all unwanted default mappers
            # we use ids so we dont accidently remove one of the previously updated default mapper
            for default_mapper in default_mappers:
                if default_mapper["id"] not in [x["id"] for x in updated_mappers]:
                    kc.delete_component(default_mapper["id"], realm)

        after_comp["mappers"] = kc.get_components(urlencode(dict(parent=cid)), realm)
        normalize_kc_comp(after_comp)
        if module._diff:
            result["diff"] = dict(before="", after=sanitize(after_comp))
        result["end_state"] = sanitize(after_comp)
        result["msg"] = f"User federation {cid} has been created"
        module.exit_json(**result)

    else:
        if state == "present":
            # Process an update

            desired_copy = deepcopy(desired_comp)
            before_copy = deepcopy(before_comp)
            # exclude bindCredential when checking wether an update is required, therefore
            # updating it only if there are other changes
            if module.params["bind_credential_update_mode"] == "only_indirect":
                desired_copy.get("config", []).pop("bindCredential", None)
                before_copy.get("config", []).pop("bindCredential", None)
            # no changes
            if desired_copy == before_copy:
                result["changed"] = False
                result["end_state"] = sanitize(desired_comp)
                result["msg"] = f"No changes required to user federation {cid}."
                module.exit_json(**result)

            # doing an update
            result["changed"] = True

            if module._diff:
                result["diff"] = dict(before=sanitize(before_comp), after=sanitize(desired_comp))

            if module.check_mode:
                module.exit_json(**result)

            # do the update
            desired_mappers = desired_comp.pop("mappers", [])
            kc.update_component(desired_comp, realm)

            for before_mapper in before_comp.get("mappers", []):
                # remove unwanted existing mappers that will not be updated
                if before_mapper["id"] not in [x["id"] for x in desired_mappers if "id" in x]:
                    kc.delete_component(before_mapper["id"], realm)

            for mapper in desired_mappers:
                if mapper in before_comp.get("mappers", []):
                    continue
                if mapper.get("id") is not None:
                    kc.update_component(mapper, realm)
                else:
                    if mapper.get("parentId") is None:
                        mapper["parentId"] = desired_comp["id"]
                    kc.create_component(mapper, realm)

            after_comp = kc.get_component(cid, realm)
            after_comp["mappers"] = sorted(
                kc.get_components(urlencode(dict(parent=cid)), realm), key=lambda x: x.get("name") or ""
            )
            normalize_kc_comp(after_comp)
            after_comp_sanitized = sanitize(after_comp)
            before_comp_sanitized = sanitize(before_comp)
            result["end_state"] = after_comp_sanitized
            if module._diff:
                result["diff"] = dict(before=before_comp_sanitized, after=after_comp_sanitized)
            result["changed"] = before_comp_sanitized != after_comp_sanitized
            result["msg"] = f"User federation {cid} has been updated"
            module.exit_json(**result)

        elif state == "absent":
            # Process a deletion
            result["changed"] = True

            if module._diff:
                result["diff"] = dict(before=sanitize(before_comp), after="")

            if module.check_mode:
                module.exit_json(**result)

            # delete it
            kc.delete_component(cid, realm)

            result["end_state"] = {}

            result["msg"] = f"User federation {cid} has been deleted"

    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021 Felix Fontein
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: hetzner_dns_record

short_description: Add or delete a single record in Hetzner DNS service

version_added: 2.0.0

description:
  - Creates and deletes single DNS records in Hetzner DNS service.
  - If you do not want to add/remove values, but replace values, you will be interested in modifying a B(record set) and not
    a single record. This is in particular important when working with C(CNAME) and C(SOA) records. Use the M(community.dns.hetzner_dns_record_set)
    module for working with record sets.
extends_documentation_fragment:
  - community.dns._hetzner
  - community.dns._hetzner.record_default_ttl
  - community.dns._hetzner.record_notes
  - community.dns._hetzner.record_type_choices
  - community.dns._hetzner.record_type_seealso
  - community.dns._hetzner.zone_id_type
  - community.dns._module_record
  - community.dns._options.record_transformation
  - community.dns._attributes
  - community.dns._attributes.actiongroup_hetzner

attributes:
  action_group:
    version_added: 2.4.0

author:
  - Markus Bergholz (@markuman) <markuman+spambelongstogoogle@gmail.com>
  - Felix Fontein (@felixfontein)
"""

EXAMPLES = r"""
- name: Add a new.foo.com A record
  community.dns.hetzner_dns_record:
    state: present
    zone: foo.com
    record: new.foo.com
    type: A
    ttl: 7200
    value: 1.1.1.1
    hetzner_token: access_token

- name: Add A record using prefix for www.example.com
  community.dns.hetzner_dns_record:
    state: present
    zone_name: example.com
    prefix: www
    type: A
    value: 198.51.100.25
    hetzner_token: "{{ lookup('env', 'HETZNER_DNS_TOKEN') }}"

- name: Remove a new.foo.com A record
  community.dns.hetzner_dns_record:
    state: absent
    zone_name: foo.com
    record: new.foo.com
    type: A
    ttl: 7200
    value: 2.2.2.2
    hetzner_token: access_token
"""

RETURN = r"""
zone_id:
  description: The ID of the zone.
  type: str
  returned: success
  sample: 23
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.dns.plugins.module_utils._argspec import (
    ModuleOptionProvider,
)
from ansible_collections.community.dns.plugins.module_utils._hetzner.api import (
    create_hetzner_api,
    create_hetzner_argument_spec,
    create_hetzner_provider_information,
)
from ansible_collections.community.dns.plugins.module_utils._http import (
    ModuleHTTPHelper,
)
from ansible_collections.community.dns.plugins.module_utils._module.record import (
    create_module_argument_spec,
    run_module,
)


def main() -> None:
    provider_information = create_hetzner_provider_information()
    argument_spec = create_hetzner_argument_spec()
    argument_spec.merge(
        create_module_argument_spec(provider_information=provider_information)
    )
    module = AnsibleModule(supports_check_mode=True, **argument_spec.to_kwargs())
    option_provider = ModuleOptionProvider(module)
    run_module(
        module,
        lambda: create_hetzner_api(option_provider, ModuleHTTPHelper(module)),
        provider_information=create_hetzner_provider_information(
            option_provider=option_provider
        ),
    )


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (C) 2018 IBM CORPORATION
# Author(s): Tzur Eliyahu <tzure@il.ibm.com>
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ibm_sa_vol
short_description: Handle volumes on IBM Spectrum Accelerate Family storage systems

description:
  - This module creates or deletes volumes to be used on IBM Spectrum Accelerate Family storage systems.
attributes:
  check_mode:
    support: none
  diff_mode:
    support: none

options:
  vol:
    description:
      - Volume name.
    required: true
    type: str
  pool:
    description:
      - Volume pool.
    type: str
  state:
    description:
      - Volume state.
    default: "present"
    choices: ["present", "absent"]
    type: str
  size:
    description:
      - Volume size.
    type: str

extends_documentation_fragment:
  - community.general._ibm_storage
  - community.general._attributes

author:
  - Tzur Eliyahu (@tzure)
"""

EXAMPLES = r"""
- name: Create a new volume.
  community.general.ibm_sa_vol:
    vol: volume_name
    pool: pool_name
    size: 17
    state: present
    username: admin
    password: secret
    endpoints: hostdev-system

- name: Delete an existing volume.
  community.general.ibm_sa_vol:
    vol: volume_name
    state: absent
    username: admin
    password: secret
    endpoints: hostdev-system
"""
RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ibm_sa_utils import (
    connect_ssl,
    execute_pyxcli_command,
    is_pyxcli_installed,
    spectrum_accelerate_spec,
)


def main():
    argument_spec = spectrum_accelerate_spec()
    argument_spec.update(
        dict(
            state=dict(default="present", choices=["present", "absent"]),
            vol=dict(required=True),
            pool=dict(),
            size=dict(),
        )
    )

    module = AnsibleModule(argument_spec)

    is_pyxcli_installed(module)

    xcli_client = connect_ssl(module)
    # required args
    volume = xcli_client.cmd.vol_list(vol=module.params.get("vol")).as_single_element
    state = module.params["state"]

    state_changed = False
    if state == "present" and not volume:
        state_changed = execute_pyxcli_command(module, "vol_create", xcli_client)
    elif state == "absent" and volume:
        state_changed = execute_pyxcli_command(module, "vol_delete", xcli_client)

    module.exit_json(changed=state_changed)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# Copyright: (c) 2019, Dag Wieers (@dagwieers) <dag@wieers.com>
# Copyright: (c) 2020, Lionel Hercot (@lhercot) <lhercot@cisco.com>
# Copyright: (c) 2021, Anvitha Jain (@anvitha-jain) <anvjain@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_site_vrf_region_cidr_subnet
short_description: Manage site-local VRF regions in schema template
description:
- Manage site-local VRF regions in schema template on Cisco ACI Multi-Site.
author:
- Dag Wieers (@dagwieers)
- Lionel Hercot (@lhercot)
- Anvitha Jain (@anvitha-jain)
options:
  schema:
    description:
    - The name of the schema.
    type: str
    required: true
  site:
    description:
    - The name of the site.
    type: str
    required: true
  template:
    description:
    - The name of the template.
    type: str
    required: true
  vrf:
    description:
    - The name of the VRF.
    type: str
    required: true
  region:
    description:
    - The name of the region.
    type: str
    required: true
  cidr:
    description:
    - The IP range of for the region CIDR.
    type: str
    required: true
  subnet:
    description:
    - The IP subnet of this region CIDR.
    type: str
    aliases: [ ip ]
  private_link_label:
    description:
    - The private link label used to represent this subnet.
    - This parameter is available for MSO version greater than 3.3.
    type: str
  zone:
    description:
    - The name of the zone for the region CIDR subnet.
    - This argument is required for AWS sites.
    type: str
    aliases: [ name ]
  vgw:
    description:
    - Whether this subnet is used for the Azure Gateway in Azure.
    - Whether this subnet is used for the Transit Gateway Attachment in AWS.
    type: bool
    aliases: [ hub_network ]
  hosted_vrf:
    description:
    - The name of hosted vrf associated with region CIDR subnet.
    - This is supported on versions of MSO that are 3.3 or greater.
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
notes:
- The ACI MultiSite PATCH API has a deficiency requiring some objects to be referenced by index.
  This can cause silent corruption on concurrent access when changing/removing on object as
  the wrong object may be referenced. This module is affected by this deficiency.
seealso:
- module: cisco.mso.mso_schema_site_vrf_region_cidr
- module: cisco.mso.mso_schema_template_vrf
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add a new site VRF region CIDR subnet
  cisco.mso.mso_schema_site_vrf_region_cidr_subnet:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    cidr: 14.14.14.1/24
    subnet: 14.14.14.2/24
    zone: us-west-1a
    state: present

- name: Remove a site VRF region CIDR subnet
  cisco.mso.mso_schema_site_vrf_region_cidr_subnet:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    cidr: 14.14.14.1/24
    subnet: 14.14.14.2/24
    state: absent

- name: Query a specific site VRF region CIDR subnet
  cisco.mso.mso_schema_site_vrf_region_cidr_subnet:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    cidr: 14.14.14.1/24
    subnet: 14.14.14.2/24
    state: query
  register: query_result

- name: Query all site VRF region CIDR subnet
  cisco.mso.mso_schema_site_vrf_region_cidr_subnet:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: Schema1
    site: Site1
    template: Template1
    vrf: VRF1
    region: us-west-1
    cidr: 14.14.14.1/24
    state: query
  register: query_result
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str", required=True),
        site=dict(type="str", required=True),
        template=dict(type="str", required=True),
        vrf=dict(type="str", required=True),
        region=dict(type="str", required=True),
        cidr=dict(type="str", required=True),
        subnet=dict(type="str", aliases=["ip"]),  # This parameter is not required for querying all objects
        private_link_label=dict(type="str"),
        zone=dict(type="str", aliases=["name"]),
        vgw=dict(type="bool", aliases=["hub_network"]),
        hosted_vrf=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["subnet"]],
            ["state", "present", ["subnet"]],
        ],
    )

    schema = module.params.get("schema")
    site = module.params.get("site")
    template = module.params.get("template").replace(" ", "")
    vrf = module.params.get("vrf")
    region = module.params.get("region")
    cidr = module.params.get("cidr")
    subnet = module.params.get("subnet")
    private_link_label = module.params.get("private_link_label")
    zone = module.params.get("zone")
    hosted_vrf = module.params.get("hosted_vrf")
    vgw = module.params.get("vgw")
    state = module.params.get("state")

    mso = MSOModule(module)

    # Get schema objects
    schema_id, schema_path, schema_obj = mso.query_schema(schema)

    # Get template
    templates = [t.get("name") for t in schema_obj.get("templates")]
    if template not in templates:
        mso.fail_json(msg="Provided template '{0}' does not exist. Existing templates: {1}".format(template, ", ".join(templates)))

    # Get site
    site_id = mso.lookup_site(site)

    # Get site_idx
    if not schema_obj.get("sites"):
        mso.fail_json(msg="No site associated with template '{0}'. Associate the site with the template using mso_schema_site.".format(template))
    sites = [(s.get("siteId"), s.get("templateName")) for s in schema_obj.get("sites")]
    sites_list = [s.get("siteId") + "/" + s.get("templateName") for s in schema_obj.get("sites")]
    if (site_id, template) not in sites:
        mso.fail_json(
            msg="Provided site/siteId/template '{0}/{1}/{2}' does not exist. "
            "Existing siteIds/templates: {3}".format(site, site_id, template, ", ".join(sites_list))
        )

    # Schema-access uses indexes
    site_idx = sites.index((site_id, template))
    # Path-based access uses site_id-template
    site_template = "{0}-{1}".format(site_id, template)

    # Get VRF at site level
    vrf_ref = mso.vrf_ref(schema_id=schema_id, template=template, vrf=vrf)
    vrfs = [v.get("vrfRef") for v in schema_obj.get("sites")[site_idx]["vrfs"]]

    # If vrf not at site level but exists at template level
    if vrf_ref not in vrfs:
        mso.fail_json(msg="Provided vrf '{0}' does not exist at site level." " Use mso_schema_site_vrf_region_cidr to create it.".format(vrf))
    vrf_idx = vrfs.index(vrf_ref)

    # Get Region
    regions = [r.get("name") for r in schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"]]
    if region not in regions:
        mso.fail_json(
            msg="Provided region '{0}' does not exist. Existing regions: {1}."
            " Use mso_schema_site_vrf_region_cidr to create it.".format(region, ", ".join(regions))
        )
    region_idx = regions.index(region)

    # Get CIDR
    cidrs = [c.get("ip") for c in schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]["cidrs"]]
    if cidr not in cidrs:
        mso.fail_json(
            msg="Provided CIDR IP '{0}' does not exist. Existing CIDR IPs: {1}."
            " Use mso_schema_site_vrf_region_cidr to create it.".format(cidr, ", ".join(cidrs))
        )
    cidr_idx = cidrs.index(cidr)

    # Get Subnet
    subnets = [s.get("ip") for s in schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]["cidrs"][cidr_idx]["subnets"]]
    if subnet is not None and subnet in subnets:
        subnet_idx = subnets.index(subnet)
        # FIXME: Changes based on index are DANGEROUS
        subnet_path = "/sites/{0}/vrfs/{1}/regions/{2}/cidrs/{3}/subnets/{4}".format(site_template, vrf, region, cidr_idx, subnet_idx)
        mso.existing = schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]["cidrs"][cidr_idx]["subnets"][subnet_idx]

    if state == "query":
        if subnet is None:
            mso.existing = schema_obj.get("sites")[site_idx]["vrfs"][vrf_idx]["regions"][region_idx]["cidrs"][cidr_idx]["subnets"]
        elif not mso.existing:
            mso.fail_json(msg="Subnet IP '{subnet}' not found".format(subnet=subnet))
        mso.exit_json()

    subnets_path = "/sites/{0}/vrfs/{1}/regions/{2}/cidrs/{3}/subnets".format(site_template, vrf, region, cidr_idx)
    ops = []

    mso.previous = mso.existing
    if state == "absent":
        if mso.existing:
            mso.sent = mso.existing = {}
            ops.append(dict(op="remove", path=subnet_path))

    elif state == "present":
        payload = dict(ip=subnet, zone="")

        if zone is not None:
            payload["zone"] = zone
        if vgw is True:
            payload["usage"] = "gateway"
        if private_link_label is not None:
            payload["privateLinkLabel"] = dict(name=private_link_label)
        if hosted_vrf is not None:
            payload["vrfRef"] = dict(schemaId=schema_id, templateName=template, vrfName=hosted_vrf)
            payload["inEditing"] = "false"

        mso.sanitize(payload, collate=True)

        if mso.existing:
            ops.append(dict(op="replace", path=subnet_path, value=mso.sent))
        else:
            ops.append(dict(op="add", path=subnets_path + "/-", value=mso.sent))

        mso.existing = mso.proposed

    if not module.check_mode:
        mso.request(schema_path, method="PATCH", data=ops)

    mso.exit_json()


if __name__ == "__main__":
    main()



# Copyright (c) 2018 Luca 'remix_tj' Lorenzetto
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

emc_vnx_argument_spec = {
    "sp_address": dict(type="str", required=True),
    "sp_user": dict(type="str", required=False, default="sysadmin"),
    "sp_password": dict(type="str", required=False, default="sysadmin", no_log=True),
}



# Copyright (c) 2018, Johannes Brunswicker <johannes.brunswicker@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    DOCUMENTATION = r"""
options:
  headers:
    description:
      - A dictionary of additional headers to be sent to POST and PUT requests.
      - Is needed for some modules.
    type: dict
    required: false
    default: {}
  utm_host:
    description:
      - The REST Endpoint of the Sophos UTM.
    type: str
    required: true
  utm_port:
    description:
      - The port of the REST interface.
    type: int
    default: 4444
  utm_token:
    description:
      - The token used to identify at the REST-API.
      - See U(https://www.sophos.com/en-us/medialibrary/PDFs/documentation/UTMonAWS/Sophos-UTM-RESTful-API.pdf?la=en), Chapter
        2.4.2.
    type: str
    required: true
  utm_protocol:
    description:
      - The protocol of the REST Endpoint.
    choices: [http, https]
    type: str
    default: https
  validate_certs:
    description:
      - Whether the REST interface's SSL certificate should be verified or not.
    type: bool
    default: true
  state:
    description:
      - The desired state of the object.
      - V(present) creates or updates an object.
      - V(absent) deletes an object if present.
    type: str
    choices: [absent, present]
    default: present
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_schema_template_contract_service_chain
version_added: "2.11.0"
short_description: Manage the Schema Template Contract Service Chaining workflow on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage the Schema Template Contract Service Chaining workflow on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.2.3) and later.
- This module is the recommended approach for Service Chaining over the previous solution using M(cisco.mso.mso_schema_template_contract_service_graph).
- This module is incompatible with M(cisco.mso.mso_schema_template_contract_service_graph).
author:
- Sabari Jaganathan (@sajagana)
options:
  schema:
    description:
    - The name of the schema.
    - This parameter is mutually exclusive with O(schema_id).
    - This parameter or O(schema_id) is required when O(template) is set.
    type: str
  schema_id:
    description:
    - The ID of the schema.
    - This parameter is mutually exclusive with O(schema).
    - This parameter or O(schema) is required when O(template) is set.
    type: str
  template:
    description:
    - The name of the template.
    - This parameter is mutually exclusive with O(template_id).
    type: str
  template_id:
    description:
    - The ID of the template.
    - This parameter is mutually exclusive with O(template).
    type: str
  contract:
    description:
    - The name of the contract.
    - This parameter is mutually exclusive with O(contract_uuid).
    type: str
  contract_uuid:
    description:
    - The UUID of the contract.
    - This parameter is mutually exclusive with O(contract).
    type: str
  node_filter:
    description:
    - The Filter After First Device option of the contract service chain.
    - This enables traffic filtering to be dynamically applied after the first device in the chain has processed the traffic.
    - Defaults to O(node_filter=allow_all) when unset during creation.
    type: str
    choices: [ allow_all, filters_from_contract ]
    aliases: [ filter_after_first_device ]
  service_nodes:
    description:
    - The list of service nodes for the contract service chain.
    - This parameter is required for creating the contract service chain.
    - Providing a new list of O(service_nodes) will completely replace an existing one from the contract service chain.
    type: list
    elements: dict
    suboptions:
      device_type:
        description:
        - The type of the service device.
        type: str
        choices: [ firewall, load_balancer, other ]
        required: true
        aliases: [ type ]
      uuid:
        description:
        - The UUID of the service device.
        - This parameter is mutually exclusive with O(service_nodes.device).
        - This parameter or O(service_nodes.device) is required.
        type: str
        aliases: [ device_uuid ]
      provider_interface_name:
        description:
        - The name of the service device interface used as the provider interface in the contract service chain.
        type: str
        required: true
        aliases: [ provider_interface, provider ]
      provider_redirect:
        description:
        - The provider redirect option of the contract service chain.
        - Defaults to O(service_nodes.provider_redirect=false) when unset during creation.
        type: bool
      consumer_interface_name:
        description:
        - The name of the service device interface used as the consumer interface in the contract service chain.
        type: str
        required: true
        aliases: [ consumer_interface, consumer ]
      consumer_redirect:
        description:
        - The consumer redirect option of the contract service chain.
        - Defaults to O(service_nodes.consumer_redirect=false) when unset during creation.
        type: bool
      device:
        description:
        - The service device details for the contract service chain.
        - This parameter is mutually exclusive with O(service_nodes.uuid).
        - This parameter or O(service_nodes.uuid) is required.
        type: dict
        suboptions:
          name:
            description:
            - The name of the service device.
            type: str
            required: true
          template:
            description:
            - The template name of the service device.
            - This parameter is mutually exclusive with O(service_nodes.device.template_id).
            - This parameter or O(service_nodes.device.template_id) is required.
            type: str
          template_id:
            description:
            - The template id of the service device.
            - This parameter is mutually exclusive with O(service_nodes.device.template).
            - This parameter or O(service_nodes.device.template) is required.
            type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
seealso:
- module: cisco.mso.mso_schema_template_contract_filter
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Add schema template contract service chain
  cisco.mso.ndo_schema_template_contract_service_chain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: schema1
    template: template1
    contract: contract1
    node_filter: allow_all
    service_nodes:
      - device_type: load_balancer
        consumer_interface_name: lb_interface1
        provider_interface_name: lb_interface2
        consumer_redirect: true
        provider_redirect: true
        device:
          name: load_balancer_device
          template: service_device_template
    state: present

- name: Update schema template contract service chain with multiple nodes
  cisco.mso.ndo_schema_template_contract_service_chain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: schema1
    template: template1
    contract: contract1
    node_filter: "filters_from_contract"
    service_nodes:
      - device_type: load_balancer
        consumer_interface_name: lb_interface1
        provider_interface_name: lb_interface2
        consumer_redirect: true
        provider_redirect: true
        device:
          name: load_balancer_device
          template: service_device_template
      - device_type: firewall
        consumer_interface_name: fw_interface1
        provider_interface_name: fw_interface2
        uuid: "{{ service_device.current.uuid }}"
    state: present

- name: Query schema template contract service chain
  cisco.mso.ndo_schema_template_contract_service_chain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: schema1
    template: template1
    contract: contract1
    state: query
  register: query_service_chain

- name: Delete schema template contract service chain
  cisco.mso.ndo_schema_template_contract_service_chain:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: schema1
    template: template1
    contract: contract1
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.templates import MSOTemplates
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate
from ansible_collections.cisco.mso.plugins.module_utils.template import KVPair
from ansible_collections.cisco.mso.plugins.module_utils.utils import snake_to_camel
from ansible_collections.cisco.mso.plugins.module_utils.constants import CONTRACT_SERVICE_CHAIN_NODE_FILTER_MAP
from ansible_collections.cisco.mso.plugins.module_utils.utils import delete_none_values
import copy


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        schema=dict(type="str"),
        schema_id=dict(type="str"),
        template=dict(type="str"),
        template_id=dict(type="str"),
        contract=dict(type="str"),
        contract_uuid=dict(type="str"),
        node_filter=dict(type="str", choices=list(CONTRACT_SERVICE_CHAIN_NODE_FILTER_MAP), aliases=["filter_after_first_device"]),
        service_nodes=dict(
            type="list",
            elements="dict",
            options=dict(
                device_type=dict(type="str", choices=["firewall", "load_balancer", "other"], required=True, aliases=["type"]),
                uuid=dict(type="str", aliases=["device_uuid"]),
                device=dict(
                    type="dict",
                    options=dict(
                        name=dict(type="str", required=True),
                        template=dict(type="str"),
                        template_id=dict(type="str"),
                    ),
                    mutually_exclusive=[["template", "template_id"]],
                    required_one_of=[["template", "template_id"]],
                ),
                provider_interface_name=dict(type="str", required=True, aliases=["provider_interface", "provider"]),
                provider_redirect=dict(type="bool"),
                consumer_interface_name=dict(type="str", required=True, aliases=["consumer_interface", "consumer"]),
                consumer_redirect=dict(type="bool"),
            ),
            mutually_exclusive=[["uuid", "device"]],
            required_one_of=[["uuid", "device"]],
        ),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["service_nodes"]],
        ],
        mutually_exclusive=[["schema", "schema_id"], ["template", "template_id"], ["contract", "contract_uuid"]],
        required_one_of=[["contract", "contract_uuid"], ["template", "template_id"]],
    )

    schema = module.params.get("schema")
    schema_id = module.params.get("schema_id")
    template = module.params.get("template")
    template_id = module.params.get("template_id")
    contract = module.params.get("contract")
    contract_uuid = module.params.get("contract_uuid")
    node_filter = CONTRACT_SERVICE_CHAIN_NODE_FILTER_MAP.get(module.params.get("node_filter"))
    service_nodes = module.params.get("service_nodes")
    state = module.params.get("state")

    mso = MSOModule(module)

    if not template_id and not (schema or schema_id):
        mso.fail_json(msg="The schema or schema_id is required when the template is set.")

    mso_templates = MSOTemplates(mso)
    mso_template = MSOTemplate(mso, "application", template, template_id, schema_name=schema, schema_id=schema_id)

    template_id = template_id or mso_template.template.get("templateId")
    template = template or mso_template.template.get("appTemplate", {}).get("template", {}).get("name")
    contract_match = mso_template.get_application_template_contract(contract_uuid, contract, fail_module=True)
    service_chain = contract_match.details.get("serviceChaining") if contract_match and contract_match.details.get("serviceChaining") else {}

    reference_collections = {
        "serviceDeviceRef": {
            "name": "deviceName",
            "reference": "deviceRef",
            "type": "serviceDevice",
            "template": "deviceTemplateName",
            "templateId": "deviceTemplateId",
        },
    }

    if service_chain:
        mso.existing = mso.previous = mso_template.update_config_with_template_and_references(
            service_chain, reference_collections, set_template=True, use_cache=True
        )  # Query a specific object

    service_chain_path = "/templates/{0}/contracts/{1}/serviceChaining".format(template, contract_match.details.get("name"))

    ops = []

    if state == "present":
        mso_values = dict(nodeFilter=node_filter)
        service_nodes_config = []
        if service_nodes:
            for index, service_node in enumerate(service_nodes, start=1):
                if not service_node.get("uuid"):
                    service_node_template = mso_templates.get_template(
                        "service_device", service_node.get("device").get("template"), service_node.get("device").get("template_id")
                    )
                    service_devices = service_node_template.template.get("deviceTemplate", {}).get("template", {}).get("devices") or []
                    service_node_match = service_node_template.get_object_by_key_value_pairs(
                        "Service Device Cluster",
                        service_devices,
                        [KVPair("name", service_node.get("device").get("name"))],
                        fail_module=True,
                    )
                    service_node["uuid"] = service_node_match.details.get("uuid")

                service_nodes_config.append(
                    {
                        "index": index,
                        "name": "node-{0}".format(index),
                        "deviceRef": service_node.get("uuid"),
                        "deviceType": snake_to_camel(service_node.get("device_type")),
                        "providerConnector": {
                            "interfaceName": service_node.get("provider_interface_name"),
                            "isRedirect": service_node.get("provider_redirect"),
                        },
                        "consumerConnector": {
                            "interfaceName": service_node.get("consumer_interface_name"),
                            "isRedirect": service_node.get("consumer_redirect"),
                        },
                    }
                )

        mso_values["serviceNodes"] = service_nodes_config
        mso.sanitize(delete_none_values(mso_values))
        ops.append(dict(op="replace" if mso.existing else "add", path=service_chain_path, value=mso.sent))

    elif state == "absent" and mso.existing:
        ops.append(dict(op="remove", path=service_chain_path))

    if not module.check_mode and ops:
        mso.request(mso_template.schema_path, method="PATCH", data=ops)
        mso_template = MSOTemplate(mso, "application", None, template_id)
        contract_match = mso_template.get_application_template_contract(contract_uuid, contract, fail_module=True)
        service_chain = contract_match.details.get("serviceChaining") if contract_match and contract_match.details.get("serviceChaining") else {}

        if service_chain:
            mso.existing = mso_template.update_config_with_template_and_references(
                service_chain, reference_collections, set_template=True, use_cache=True
            )  # When the state is present
        else:
            mso.existing = {}  # When the state is absent

    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        if state == "present":
            mso.existing = mso_template.update_config_with_template_and_references(
                copy.deepcopy(mso.proposed), reference_collections, set_template=True, use_cache=True
            )
        else:
            mso.existing = {}

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_info
short_description: Gather information about Intersight
description:
- Gathers information about servers in L(Cisco Intersight,https://intersight.com).
- This module was called C(intersight_facts) before Ansible 2.9. The usage did not change.
extends_documentation_fragment: intersight
options:
  server_names:
    description:
    - Server names to retrieve information from.
    - An empty list will return all servers.
    type: list
    elements: str
author:
- David Soper (@dsoper2)
- CiscoUcs (@CiscoUcs)
'''

EXAMPLES = r'''
- name: Get info for all servers
  intersight_info:
    api_private_key: ~/Downloads/SecretKey.txt
    api_key_id: 64612d300d0982/64612d300d0b00/64612d300d3650
    server_names:
- debug:
    msg: "server name {{ item.Name }}, moid {{ item.Moid }}"
  loop: "{{ intersight_servers }}"
  when: intersight_servers is defined

- name: Get info for servers by name
  intersight_info:
    api_private_key: ~/Downloads/SecretKey.txt
    api_key_id: 64612d300d0982/64612d300d0b00/64612d300d3650
    server_names:
      - SJC18-L14-UCS1-1
- debug:
    msg: "server moid {{ intersight_servers[0].Moid }}"
  when: intersight_servers[0] is defined
'''

RETURN = r'''
intersight_servers:
  description: A list of Intersight Servers.  See L(Cisco Intersight,https://intersight.com/apidocs) for details.
  returned: always
  type: complex
  contains:
    Name:
      description: The name of the server.
      returned: always
      type: str
      sample: SJC18-L14-UCS1-1
    Moid:
      description: The unique identifier of this Managed Object instance.
      returned: always
      type: str
      sample: 5978bea36ad4b000018d63dc
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def get_servers(module, intersight):
    query_list = []
    if module.params['server_names']:
        for server in module.params['server_names']:
            query_list.append("Name eq '%s'" % server)
    query_str = ' or '.join(query_list)
    options = {
        'http_method': 'get',
        'resource_path': '/compute/PhysicalSummaries',
        'query_params': {
            '$filter': query_str,
            '$top': 1000,
        }
    }
    response_dict = intersight.call_api(**options)

    return response_dict.get('Results')


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        server_names=dict(type='list', elements='str'),
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)

    # one API call returning all requested servers
    module.exit_json(intersight_servers=get_servers(module, intersight))


if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_serviceendpointpolicy_info
version_added: "3.9.0"
short_description: Get or list the service endpoint policy
description:
    - Get or list the service endpoint policy.

options:
    resource_group:
        description:
            - The name of the resource group.
        type: str
    name:
        description:
            - The name of the service endpoint policy name.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - magodo (@magodo)
    - Fred Sun (@Fred-sun)
'''

EXAMPLES = '''
- name: Get the specified service endpoint policy
  azure.azcollection.azure_rm_serviceendpointpolicy_info:
    resource_group: testRG
    name: testpolicy

- name: List all service endpoint policy under same resource group
  azure.azcollection.azure_rm_serviceendpointpolicy_info:
    resource_group: testRG

- name: List all service endpoint policy
  azure.azcollection.azure_rm_serviceendpointpolicy_info:
'''

RETURN = '''
serviceendpointpolicies:
    description:
        - List the facts of service endpoint policy.
    type: complex
    returned: always
    contains:
        id:
            description:
                - Fully qualified ID of the resource.
            type: str
            returned: always
            sample: "/subscriptions/xxx-xxx/resourceGroups/TestRG/providers/Microsoft.Network/serviceEndpointPolicies/testpolicy"
        name:
            description:
                - The name of the resource that is unique within a resource group.
            type: str
            returned: always
            sample: testpolicy
        etag:
            description:
                - A unique read-only string that changes whenever the resource is updated.
            returned: always
            type: str
            sample: "49e15a7b-1bff-46ba-b962-7a11e75cb729"
        resource_group:
            description:
                - The name of the resource group.
            type: str
            returned: always
            sample: TestRG
        location:
            description:
                - The resource's location.
            type: str
            returned: always
            sample: eastus
        type:
            description:
                - The type of the resource.
            type: str
            returned: always
            sample: Microsoft.Network/serviceEndpointPolicies
        service_endpoint_policy_definitions:
            description:
                - A collection of service endpoint policy definitions of the service endpoint policy.
            type: list
            returned: always
            sample: []
        provisioning_state:
            description:
                - The provisioning state of the service endpoint policy resource.
            type: str
            returned: always
            sample: Successded
        resource_guid:
            description:
                - The resource GUID property of the service endpoint policy resource.
            type: str
            returned: always
            sample: 77ee69ee-4152-4271-bfc0-1d431b7b7d28
'''

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase


class AzureRMServiceEndpointPolicyInfo(AzureRMModuleBase):
    """Information class for an Azure RM Data Collection Rules"""

    def __init__(self):
        self.module_arg_spec = dict(
            resource_group=dict(type='str'),
            name=dict(type='str'),
            tags=dict(type='list', elements='str')
        )

        self.resource_group = None
        self.name = None
        self.tags = None
        self.log_path = None
        self.log_mode = None
        self.required_by = {
            'name': 'resource_group'
        }

        self.results = dict(
            changed=False,
            serviceendpointpolicies=[]
        )

        super(AzureRMServiceEndpointPolicyInfo, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                               supports_check_mode=True,
                                                               supports_tags=False,
                                                               facts_module=True,
                                                               required_by=self.required_by)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        result = []
        if self.name:
            result = self.get_policy(self.name)
        else:
            result = self.list_all_policy()
        self.results['serviceendpointpolicies'] = result

        return self.results

    def get_policy(self, name):
        '''
        Gets the specified service endpoint policy
        '''
        result = []
        response = None

        try:
            response = self.network_client.service_endpoint_policies.get(self.resource_group, self.name)
        except Exception:
            self.log("Could not get the specified service endpoint policy {0}".format(self.name))
            return []
        if response and self.has_tags(response.tags, self.tags):
            result = [self.format_item(response)]
        return result

    def list_all_policy(self):
        '''
        List all service endpoint policys for the specified resource.
        '''
        result = []
        response = None
        try:
            if self.resource_group:
                response = self.network_client.service_endpoint_policies.list_by_resource_group(self.resource_group)
            else:
                response = self.network_client.service_endpoint_policies.list()
        except Exception as ex:
            self.log("Could not list all service endpoint policys for the specified resource, Exception as {0}".format(ex))
            return []
        if response:
            for item in response:
                if self.has_tags(item.tags, self.tags):
                    result.append(self.format_item(item))
        return result

    def format_item(self, item):
        response = item.as_dict()
        response['resource_group'] = self.parse_resource_to_dict(response['id']).get('resource_group')
        return response


def main():
    """Main execution"""
    AzureRMServiceEndpointPolicyInfo()


if __name__ == '__main__':
    main()



# The contents of this file are licensed under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with the
# License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

base_interfaces = {
    "ATM": "ATM",
    "AT": "ATM",
    "B": "Bdi",
    "Bd": "Bdi",
    "Bdi": "Bdi",
    "EOBC": "EOBC",
    "EO": "EOBC",
    "Ethernet": "Ethernet",
    "Eth": "Ethernet",
    "eth": "Ethernet",
    "Et": "Ethernet",
    "et": "Ethernet",
    "FastEthernet": "FastEthernet",
    "FastEth": "FastEthernet",
    "FastE": "FastEthernet",
    "Fast": "FastEthernet",
    "Fas": "FastEthernet",
    "FE": "FastEthernet",
    "Fa": "FastEthernet",
    "fa": "FastEthernet",
    "Fddi": "Fddi",
    "FD": "Fddi",
    "FortyGigabitEthernet": "FortyGigabitEthernet",
    "FortyGigEthernet": "FortyGigabitEthernet",
    "FortyGigEth": "FortyGigabitEthernet",
    "FortyGigE": "FortyGigabitEthernet",
    "FortyGig": "FortyGigabitEthernet",
    "FGE": "FortyGigabitEthernet",
    "FO": "FortyGigabitEthernet",
    "Fo": "FortyGigabitEthernet",
    "FiftyGigabitEthernet": "FiftyGigabitEthernet",
    "FiftyGigEthernet": "FiftyGigabitEthernet",
    "FiftyGigEth": "FiftyGigabitEthernet",
    "FiftyGigE": "FiftyGigabitEthernet",
    "FI": "FiftyGigabitEthernet",
    "Fi": "FiftyGigabitEthernet",
    "fi": "FiftyGigabitEthernet",
    "GigabitEthernet": "GigabitEthernet",
    "GigEthernet": "GigabitEthernet",
    "GigEth": "GigabitEthernet",
    "GigE": "GigabitEthernet",
    "Gig": "GigabitEthernet",
    "GE": "GigabitEthernet",
    "Ge": "GigabitEthernet",
    "ge": "GigabitEthernet",
    "Gi": "GigabitEthernet",
    "gi": "GigabitEthernet",
    "HundredGigabitEthernet": "HundredGigabitEthernet",
    "HundredGigEthernet": "HundredGigabitEthernet",
    "HundredGigEth": "HundredGigabitEthernet",
    "HundredGigE": "HundredGigabitEthernet",
    "HundredGig": "HundredGigabitEthernet",
    "Hu": "HundredGigabitEthernet",
    "TwentyFiveGigabitEthernet": "TwentyFiveGigabitEthernet",
    "TwentyFiveGigEthernet": "TwentyFiveGigabitEthernet",
    "TwentyFiveGigEth": "TwentyFiveGigabitEthernet",
    "TwentyFiveGigE": "TwentyFiveGigabitEthernet",
    "TwentyFiveGig": "TwentyFiveGigabitEthernet",
    "TF": "TwentyFiveGigabitEthernet",
    "Tf": "TwentyFiveGigabitEthernet",
    "tf": "TwentyFiveGigabitEthernet",
    "TwoHundredGigabitEthernet": "TwoHundredGigabitEthernet",
    "TwoHundredGigEthernet": "TwoHundredGigabitEthernet",
    "TwoHundredGigEth": "TwoHundredGigabitEthernet",
    "TwoHundredGigE": "TwoHundredGigabitEthernet",
    "TwoHundredGig": "TwoHundredGigabitEthernet",
    "TH": "TwoHundredGigabitEthernet",
    "Th": "TwoHundredGigabitEthernet",
    "th": "TwoHundredGigabitEthernet",
    "FourHundredGigabitEthernet": "FourHundredGigabitEthernet",
    "FourHundredGigEthernet": "FourHundredGigabitEthernet",
    "FourHundredGigEth": "FourHundredGigabitEthernet",
    "FourHundredGigE": "FourHundredGigabitEthernet",
    "FourHundredGig": "FourHundredGigabitEthernet",
    "F": "FourHundredGigabitEthernet",
    "f": "FourHundredGigabitEthernet",
    "Loopback": "Loopback",
    "loopback": "Loopback",
    "Lo": "Loopback",
    "lo": "Loopback",
    "Management": "Management",
    "Mgmt": "Management",
    "mgmt": "Management",
    "Ma": "Management",
    "Management_short": "Ma",
    "MFR": "MFR",
    "Multilink": "Multilink",
    "Mu": "Multilink",
    "n": "nve",
    "nv": "nve",
    "nve": "nve",
    "PortChannel": "Port-channel",
    "Port-channel": "Port-channel",
    "Port-Channel": "Port-channel",
    "port-channel": "Port-channel",
    "po": "Port-channel",
    "Po": "Port-channel",
    "POS": "POS",
    "PO": "POS",
    "Serial": "Serial",
    "Se": "Serial",
    "S": "Serial",
    "TenGigabitEthernet": "TenGigabitEthernet",
    "TenGigEthernet": "TenGigabitEthernet",
    "TenGigEth": "TenGigabitEthernet",
    "TenGig": "TenGigabitEthernet",
    "TeGig": "TenGigabitEthernet",
    "Ten": "TenGigabitEthernet",
    "T": "TenGigabitEthernet",
    "Te": "TenGigabitEthernet",
    "te": "TenGigabitEthernet",
    "Tunnel": "Tunnel",
    "Tun": "Tunnel",
    "Tu": "Tunnel",
    "Twe": "TwentyFiveGigE",
    "Tw": "TwoGigabitEthernet",
    "Two": "TwoGigabitEthernet",
    "Virtual-Access": "Virtual-Access",
    "Vi": "Virtual-Access",
    "Virtual-Template": "Virtual-Template",
    "Vt": "Virtual-Template",
    "VLAN": "VLAN",
    "V": "VLAN",
    "Vl": "VLAN",
    "Wlan-GigabitEthernet": "Wlan-GigabitEthernet",
}

reverse_mapping = {
    "ATM": "At",
    "EOBC": "EO",
    "Ethernet": "Et",
    "FastEthernet": "Fa",
    "Fddi": "FD",
    "FortyGigabitEthernet": "Fo",
    "GigabitEthernet": "Gi",
    "HundredGigabitEthernet": "Hu",
    "Loopback": "Lo",
    "Management": "Ma",
    "MFR": "MFR",
    "Multilink": "Mu",
    "Port-channel": "Po",
    "POS": "PO",
    "Serial": "Se",
    "TenGigabitEthernet": "Te",
    "Tunnel": "Tu",
    "TwoGigabitEthernet": "Two",
    "TwentyFiveGigE": "Twe",
    "Virtual-Access": "Vi",
    "Virtual-Template": "Vt",
    "VLAN": "Vl",
    "Wlan-GigabitEthernet": "Wl-Gi",
}



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Tim Cragg (@timcragg)
# Copyright: (c) 2023, Akini Ross (@akinross)
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: aci_dhcp_relay_provider
short_description: Manage DHCP relay policy providers (dhcp:RsProv)
description:
- Manage DHCP relay policy providers configuration on Cisco ACI fabrics.
options:
  tenant:
    description:
    - Name of the tenant the relay_policy is in.
    - When tenant is not provided the module will be applied to the global (infra) policy.
    type: str
  relay_policy:
    description:
    - Name of an existing DHCP relay policy
    type: str
    aliases: [ relay_policy_name ]
  provider_tenant:
    description:
    - Name of the tenant the epg or external_epg is in
    - Required when epg or external_epg is in a different tenant than the relay_policy
    - Required when global (infra) relay_policy is configured with epg or external_epg types
    type: str
  epg_type:
    description:
    - Type of EPG the DHCP server is in
    type: str
    choices: [ epg, l2_external, l3_external, dn ]
  anp:
    description:
    - Application Profile the EPG is in.
    - Only used when epg_type is app_epg.
    type: str
  epg:
    description:
    - Name of the Application EPG the DHCP server is in.
    - Only used when epg_type is epg
    type: str
    aliases: [ app_epg ]
  l2out_name:
    description:
    - Name of the L2out the DHCP server is in.
    - Only used when epg_type is l2_external
    type: str
  l3out_name:
    description:
    - Name of the L3out the DHCP server is in.
    - Only used when epg_type is l3_external.
    type: str
  external_epg:
    description:
    - Name of the external network object the DHCP server is in.
    - Only used when epg_type is l2_external or l3_external.
    type: str
    aliases: [ external_net ]
  dn:
    description:
    - dn of the EPG the DHCP server is in
    - Only used when epg_type is dn
    type: str
  dhcp_server_addr:
    description:
    - DHCP server address
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant) and C(relay_policy) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant) and C(cisco.aci.aci_dhcp_relay) modules can be used for this.
seealso:
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(dhcp:RsProv).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Tim Cragg (@timcragg)
- Akini Ross (@akinross)
"""

EXAMPLES = r"""
- name: Add a new DHCP relay App EPG provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: Auto-Demo
    relay_policy: my_dhcp_relay
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    dhcp_server_addr: 10.20.30.40
    state: present
  delegate_to: localhost

- name: Add a new Global (infra) DHCP relay App EPG provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    relay_policy: my_dhcp_relay
    provider_tenant: Auto-Demo
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    dhcp_server_addr: 10.20.30.40
    state: present
  delegate_to: localhost

- name: Add a new DHCP relay L3out provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: Auto-Demo
    relay_policy: my_dhcp_relay
    epg_type: l3_external
    l3out_name: my_l3out
    external_net: my_l3out_ext_net
    dhcp_server_addr: 10.20.30.40
    state: present
  delegate_to: localhost

- name: Remove a DHCP relay provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: Auto-Demo
    relay_policy: my_dhcp_relay
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    state: absent
  delegate_to: localhost

- name: Remove a Global (infra) DHCP relay provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    relay_policy: my_dhcp_relay
    provider_tenant: Auto-Demo
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    state: absent
  delegate_to: localhost

- name: Query a DHCP relay provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: Auto-Demo
    relay_policy: my_dhcp_relay
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    state: query
  delegate_to: localhost
  register: query_result

- name: Query a Global (infra) DHCP relay provider
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    relay_policy: my_dhcp_relay
    provider_tenant: Auto-Demo
    epg_type: epg
    anp: my_anp
    epg: my_app_epg
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all DHCP relay providers in a specific tenant
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: Auto-Demo
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all DHCP relay providers
  cisco.aci.aci_dhcp_relay_provider:
    host: apic
    username: admin
    password: SomeSecretPassword
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
   current:
     description: The existing configuration from the APIC after the module has finished
     returned: success
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production environment",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   error:
     description: The error information as returned from the APIC
     returned: failure
     type: dict
     sample:
       {
           "code": "122",
           "text": "unknown managed object class foo"
       }
   raw:
     description: The raw output returned by the APIC REST API (xml or json)
     returned: parse error
     type: str
     sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class "/></imdata>'
   sent:
     description: The actual/minimal configuration pushed to the APIC
     returned: info
     type: list
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment"
               }
           }
       }
   previous:
     description: The original configuration from the APIC before the module has started
     returned: info
     type: list
     sample:
       [
           {
               "fvTenant": {
                   "attributes": {
                       "descr": "Production",
                       "dn": "uni/tn-production",
                       "name": "production",
                       "nameAlias": "",
                       "ownerKey": "",
                       "ownerTag": ""
                   }
               }
           }
       ]
   proposed:
     description: The assembled configuration from the user-provided parameters
     returned: info
     type: dict
     sample:
       {
           "fvTenant": {
               "attributes": {
                   "descr": "Production environment",
                   "name": "production"
               }
           }
       }
   filter_string:
     description: The filter string used for the request
     returned: failure or debug
     type: str
     sample: ?rsp-prop-include=config-only
   method:
     description: The HTTP method used for the request to the APIC
     returned: failure or debug
     type: str
     sample: POST
   response:
     description: The HTTP response from the APIC
     returned: failure or debug
     type: str
     sample: OK (30 bytes)
   status:
     description: The HTTP status from the APIC
     returned: failure or debug
     type: int
     sample: 200
   url:
     description: The HTTP url used for the request to the APIC
     returned: failure or debug
     type: str
     sample: https://10.11.12.13/api/mo/uni/tn-production.json
   """

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        relay_policy=dict(type="str", aliases=["relay_policy_name"]),
        epg_type=dict(type="str", choices=["epg", "l2_external", "l3_external", "dn"]),
        anp=dict(type="str"),
        epg=dict(type="str", aliases=["app_epg"]),
        l2out_name=dict(type="str"),
        l3out_name=dict(type="str"),
        external_epg=dict(type="str", aliases=["external_net"]),
        dhcp_server_addr=dict(type="str"),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        tenant=dict(type="str"),
        provider_tenant=dict(type="str"),
        dn=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["relay_policy", "epg_type"]],
            ["state", "present", ["relay_policy", "epg_type"]],
            ["epg_type", "epg", ["anp", "epg"]],
            ["epg_type", "l2_external", ["l2out_name", "external_epg"]],
            ["epg_type", "l3_external", ["l3out_name", "external_epg"]],
            ["epg_type", "l3_external", ["provider_tenant", "tenant"], True],
            ["epg_type", "l2_external", ["provider_tenant", "tenant"], True],
            ["epg_type", "epg", ["provider_tenant", "tenant"], True],
            ["epg_type", "dn", ["dn"]],
        ],
        mutually_exclusive=[
            ["anp", "l2out_name"],
            ["anp", "l3out_name"],
            ["anp", "external_epg"],
            ["anp", "dn"],
            ["epg", "l2out_name"],
            ["epg", "l3out_name"],
            ["epg", "external_epg"],
            ["epg", "dn"],
            ["l2out_name", "l3out_name"],
            ["l2out_name", "dn"],
            ["l3out_name", "dn"],
            ["external_epg", "dn"],
        ],
    )

    relay_policy = module.params.get("relay_policy")
    state = module.params.get("state")
    tenant = module.params.get("tenant")
    epg_type = module.params.get("epg_type")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    l2out_name = module.params.get("l2out_name")
    l3out_name = module.params.get("l3out_name")
    external_epg = module.params.get("external_epg")
    dhcp_server_addr = module.params.get("dhcp_server_addr")
    provider_tenant = module.params.get("provider_tenant")
    dn = module.params.get("dn")

    if provider_tenant is None:
        provider_tenant = tenant

    if epg_type == "epg":
        tdn = "uni/tn-{0}/ap-{1}/epg-{2}".format(provider_tenant, anp, epg)
    elif epg_type == "l2_external":
        tdn = "uni/tn-{0}/l2out-{1}/instP-{2}".format(provider_tenant, l2out_name, external_epg)
    elif epg_type == "l3_external":
        tdn = "uni/tn-{0}/out-{1}/instP-{2}".format(provider_tenant, l3out_name, external_epg)
    elif epg_type == "dn":
        tdn = dn
    else:
        tdn = None

    if tenant is None:
        root_class = dict(
            aci_class="dhcpRelayP",
            aci_rn="infra/relayp-{0}".format(relay_policy),
            module_object=relay_policy,
            target_filter={"name": relay_policy},
        )
        subclass_1 = dict(
            aci_class="dhcpRsProv",
            aci_rn="rsprov-[{0}]".format(tdn),
            module_object=tdn,
            target_filter={"tDn": tdn},
        )
        subclass_2 = None
    else:
        root_class = dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        )
        subclass_1 = dict(
            aci_class="dhcpRelayP",
            aci_rn="relayp-{0}".format(relay_policy),
            module_object=relay_policy,
            target_filter={"name": relay_policy},
        )
        subclass_2 = dict(
            aci_class="dhcpRsProv",
            aci_rn="rsprov-[{0}]".format(tdn),
            module_object=tdn,
            target_filter={"tDn": tdn},
        )

    aci = ACIModule(module)

    aci.construct_url(root_class=root_class, subclass_1=subclass_1, subclass_2=subclass_2)

    aci.get_existing()

    if state == "present":
        aci.payload(
            aci_class="dhcpRsProv",
            class_config=dict(addr=dhcp_server_addr, tDn=tdn),
        )

        aci.get_diff(aci_class="dhcpRsProv")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) 2021, Roberto Moreda <moreda@allenta.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
module: dnf_versionlock
version_added: '4.0.0'
short_description: Locks package versions in C(dnf) based systems
description:
  - Locks package versions using the C(versionlock) plugin in C(dnf) based systems. This plugin takes a set of name and versions
    for packages and excludes all other versions of those packages. This allows you to for example protect packages from being
    updated by newer versions. The state of the plugin that reflects locking of packages is the C(locklist).
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: partial
    details:
      - The logics of the C(versionlock) plugin for corner cases could be confusing, so please take in account that this module
        does its best to give a C(check_mode) prediction on what is going to happen. In case of doubt, check the documentation
        of the plugin.
      - Sometimes the module could predict changes in C(check_mode) that are not fulfilled because C(versionlock) concludes
        that there is already a entry in C(locklist) that already matches.
  diff_mode:
    support: none
options:
  name:
    description:
      - Package name spec to add or exclude to or delete from the C(locklist) using the format expected by the C(dnf repoquery)
        command.
      - This parameter is mutually exclusive with O(state=clean).
    type: list
    elements: str
    default: []
  raw:
    description:
      - Do not resolve package name specs to NEVRAs to find specific version to lock to. Instead the package name specs are
        used as they are. This enables locking to not yet available versions of the package.
    type: bool
    default: false
  state:
    description:
      - Whether to add (V(present) or V(excluded)) to or remove (V(absent) or V(clean)) from the C(locklist).
      - V(present) adds a package name spec to the C(locklist). If there is a installed package that matches, then only that
        version is added. Otherwise, all available package versions are added.
      - V(excluded) adds a package name spec as excluded to the C(locklist). It means that packages represented by the package
        name spec are excluded from transaction operations. All available package versions are added.
      - V(absent) deletes entries in the C(locklist) that match the package name spec.
      - V(clean) deletes all entries in the C(locklist). This option is mutually exclusive with O(name).
    choices: ['absent', 'clean', 'excluded', 'present']
    type: str
    default: present
notes:
  - In an ideal world, the C(versionlock) plugin would have a dry-run option to know for sure what is going to happen. So
    far we have to work with a best guess as close as possible to the behaviour inferred from its code.
  - For most of cases where you want to lock and unlock specific versions of a package, this works fairly well.
  - Does not work with C(dnf5).
requirements:
  - dnf
  - dnf-plugin-versionlock
author:
  - Roberto Moreda (@moreda) <moreda@allenta.com>
"""

EXAMPLES = r"""
- name: Prevent installed nginx from being updated
  community.general.dnf_versionlock:
    name: nginx
    state: present

- name: Prevent multiple packages from being updated
  community.general.dnf_versionlock:
    name:
      - nginx
      - haproxy
    state: present

- name: Remove lock from nginx to be updated again
  community.general.dnf_versionlock:
    name: nginx
    state: absent

- name: Exclude bind 32:9.11 from installs or updates
  community.general.dnf_versionlock:
    name: bind-32:9.11*
    state: excluded

- name: Keep bash package in major version 4
  community.general.dnf_versionlock:
    name: bash-0:4.*
    raw: true
    state: present

- name: Delete all entries in the locklist of versionlock
  community.general.dnf_versionlock:
    state: clean
"""

RETURN = r"""
locklist_pre:
  description: Locklist before module execution.
  returned: success
  type: list
  elements: str
  sample: ["bash-0:4.4.20-1.el8_4.*", "!bind-32:9.11.26-4.el8_4.*"]
locklist_post:
  description: Locklist after module execution.
  returned: success and (not check mode or state is clean)
  type: list
  elements: str
  sample: ["bash-0:4.4.20-1.el8_4.*"]
specs_toadd:
  description: Package name specs meant to be added by versionlock.
  returned: success
  type: list
  elements: str
  sample: ["bash"]
specs_todelete:
  description: Package name specs meant to be deleted by versionlock.
  returned: success
  type: list
  elements: str
  sample: ["bind"]
"""

import fnmatch
import os
import re

from ansible.module_utils.basic import AnsibleModule

DNF_BIN = "/usr/bin/dnf"
VERSIONLOCK_CONF = "/etc/dnf/plugins/versionlock.conf"
# NEVRA regex.
NEVRA_RE = re.compile(r"^(?P<name>.+)-(?P<epoch>\d+):(?P<version>.+)-(?P<release>.+)\.(?P<arch>.+)$")


def do_versionlock(module, command, patterns=None, raw=False):
    patterns = patterns if patterns else []
    raw_parameter = ["--raw"] if raw else []
    # Call dnf versionlock using a just one full NEVR package-name-spec each
    # time because multiple package-name-spec and globs are not well supported.
    #
    # This is a workaround for two alleged bugs in the dnf versionlock plugin:
    # * Multiple package-name-spec arguments don't lock correctly
    #   (https://bugzilla.redhat.com/show_bug.cgi?id=2013324).
    # * Locking a version of a not installed package disallows locking other
    #   versions later (https://bugzilla.redhat.com/show_bug.cgi?id=2013332)
    #
    # NOTE: This is suboptimal in terms of performance if there are more than a
    # few package-name-spec patterns to lock, because there is a command
    # execution per each. This will improve by changing the strategy once the
    # mentioned alleged bugs in the dnf versionlock plugin are fixed.
    if patterns:
        outs = []
        for p in patterns:
            rc, out, err = module.run_command(
                [DNF_BIN, "-q", "versionlock", command] + raw_parameter + [p], check_rc=True
            )
            outs.append(out)
        out = "\n".join(outs)
    else:
        rc, out, err = module.run_command([DNF_BIN, "-q", "versionlock", command], check_rc=True)
    return out


# This is equivalent to the _match function of the versionlock plugin.
def match(entry, pattern):
    entry = entry.lstrip("!")
    if entry == pattern:
        return True
    m = NEVRA_RE.match(entry)
    if not m:
        return False

    for name in (
        f"{m['name']}",
        f"{m['name']}.{m['arch']}",
        f"{m['name']}-{m['version']}",
        f"{m['name']}-{m['version']}-{m['release']}",
        f"{m['name']}-{m['epoch']}:{m['version']}",
        f"{m['name']}-{m['version']}-{m['release']}.{m['arch']}",
        f"{m['name']}-{m['epoch']}:{m['version']}-{m['release']}",
        f"{m['epoch']}:{m['name']}-{m['version']}-{m['release']}.{m['arch']}",
        f"{m['name']}-{m['epoch']}:{m['version']}-{m['release']}.{m['arch']}",
    ):
        if fnmatch.fnmatch(name, pattern):
            return True
    return False


def get_packages(module, patterns, only_installed=False):
    packages_available_map_name_evrs = {}
    rc, out, err = module.run_command(
        [DNF_BIN, "-q", "repoquery"] + (["--installed"] if only_installed else []) + patterns, check_rc=True
    )

    for p in out.split():
        # Extract the NEVRA pattern.
        m = NEVRA_RE.match(p)
        if not m:
            module.fail_json(msg=f"failed to parse nevra for {p}", rc=rc, out=out, err=err)

        evr = f"{m['epoch']}:{m['version']}-{m['release']}"

        packages_available_map_name_evrs.setdefault(m["name"], set())
        packages_available_map_name_evrs[m["name"]].add(evr)
    return packages_available_map_name_evrs


def get_package_mgr():
    for bin_path in (DNF_BIN,):
        if os.path.exists(bin_path):
            return "dnf5" if os.path.realpath(bin_path) == "/usr/bin/dnf5" else "dnf"
    # fallback to dnf
    return "dnf"


def get_package_list(module, package_mgr="dnf"):
    if package_mgr == "dnf":
        return do_versionlock(module, "list").split()

    package_list = []
    if package_mgr == "dnf5":
        stanza_start = False
        package_name = None
        for line in do_versionlock(module, "list").splitlines():
            if line.startswith(("#", " ")):
                continue
            if line.startswith("Package name:"):
                stanza_start = True
                dummy, name = line.split(":", 1)
                name = name.strip()
                pkg_name = get_packages(module, patterns=[name])
                package_name = f"{name}-{pkg_name[name].pop()}.*"
                if package_name and package_name not in package_list:
                    package_list.append(package_name)
            if line.startswith("evr"):
                dummy, package_version = line.split("=", 1)
                package_version = package_version.strip()
                if stanza_start:
                    if package_name and package_name not in package_list:
                        package_list.append(package_name)
                    stanza_start = False
    return package_list


def main():
    module = AnsibleModule(
        argument_spec=dict(
            name=dict(type="list", elements="str", default=[]),
            raw=dict(type="bool", default=False),
            state=dict(type="str", default="present", choices=["present", "absent", "excluded", "clean"]),
        ),
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    patterns = module.params["name"]
    raw = module.params["raw"]
    state = module.params["state"]
    changed = False
    msg = ""

    # Check module pre-requisites.
    global DNF_BIN
    DNF_BIN = module.get_bin_path("dnf", True)
    package_mgr = get_package_mgr()
    if package_mgr == "dnf" and not os.path.exists(VERSIONLOCK_CONF):
        module.fail_json(msg="plugin versionlock is required")

    # Check incompatible options.
    if state == "clean" and patterns:
        module.fail_json(msg="clean state is incompatible with a name list")
    if state != "clean" and not patterns:
        module.fail_json(msg=f"name list is required for {state} state")

    locklist_pre = get_package_list(module, package_mgr=package_mgr)

    specs_toadd = []
    specs_todelete = []

    if state in ["present", "excluded"]:
        if raw:
            # Add raw patterns as specs to add.
            for p in patterns:
                if (p if state == "present" else f"!{p}") not in locklist_pre:
                    specs_toadd.append(p)
        else:
            # Get available packages that match the patterns.
            packages_map_name_evrs = get_packages(module, patterns)

            # Get installed packages that match the patterns.
            packages_installed_map_name_evrs = get_packages(module, patterns, only_installed=True)

            # Obtain the list of package specs that require an entry in the
            # locklist. This list is composed by:
            #  a) the non-installed packages list with all available
            #     versions
            #  b) the installed packages list
            packages_map_name_evrs.update(packages_installed_map_name_evrs)
            for name in packages_map_name_evrs:
                for evr in packages_map_name_evrs[name]:
                    locklist_entry = f"{name}-{evr}.*"

                    if (locklist_entry if state == "present" else f"!{locklist_entry}") not in locklist_pre:
                        specs_toadd.append(locklist_entry)

        if specs_toadd and not module.check_mode:
            cmd = "add" if state == "present" else "exclude"
            msg = do_versionlock(module, cmd, patterns=specs_toadd, raw=raw)

    elif state == "absent":
        if raw:
            # Add raw patterns as specs to delete.
            for p in patterns:
                if p in locklist_pre:
                    specs_todelete.append(p)

        else:
            # Get patterns that match the some line in the locklist.
            for p in patterns:
                for e in locklist_pre:
                    if match(e, p):
                        specs_todelete.append(p)

        if specs_todelete and not module.check_mode:
            msg = do_versionlock(module, "delete", patterns=specs_todelete, raw=raw)

    elif state == "clean":
        specs_todelete = locklist_pre

        if specs_todelete and not module.check_mode:
            msg = do_versionlock(module, "clear")

    if specs_toadd or specs_todelete:
        changed = True

    response = {
        "changed": changed,
        "msg": msg,
        "locklist_pre": locklist_pre,
        "specs_toadd": specs_toadd,
        "specs_todelete": specs_todelete,
    }
    if not module.check_mode:
        response["locklist_post"] = get_package_list(module, package_mgr=package_mgr)
    else:
        if state == "clean":
            response["locklist_post"] = []

    module.exit_json(**response)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_system_qos_policy
short_description: System QoS policy configuration for Cisco Intersight
description:
  - System QoS policy configuration for Cisco Intersight.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs).
extends_documentation_fragment: intersight
options:
  state:
    description:
      - If C(present), will verify the resource is present and will create if needed.
      - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
  organization:
    description:
      - The name of the Organization this resource is assigned to.
      - Profiles and Policies that are created within a Custom Organization are applicable only to devices in the same Organization.
    type: str
    default: default
  name:
    description:
      - The name assigned to the System QoS policy.
      - The name must be between 1 and 64 alphanumeric characters, allowing special characters :-_.
    type: str
    required: true
  tags:
    description:
      - List of tags in Key:<user-defined key> Value:<user-defined value> format.
    type: list
    elements: dict
  description:
    description:
      - The user-defined description of the System QoS policy.
      - Description can contain letters(a-z, A-Z), numbers(0-9), hyphen(-), period(.), colon(:), or an underscore(_).
    type: str
    aliases: [descr]
  classes:
    description:
      - List of QoS classes configured in the policy.
      - System includes predefined classes (Bronze, Silver, Gold, Platinum, Best Effort, FC).
      - Best Effort and FC classes are always enabled and have restrictions on modifiable parameters.
    type: list
    elements: dict
    suboptions:
      name:
        description:
          - Name of the QoS class.
          - Valid class names are Bronze, Silver, Gold, Platinum, Best Effort, FC.
        type: str
        required: true
        choices: [Bronze, Silver, Gold, Platinum, "Best Effort", FC]
      admin_state:
        description:
          - Administrative state of the QoS class.
          - Best Effort and FC classes are always enabled.
          - Bronze, Silver, Gold, and Platinum classes are disabled by default when not specified.
          - If not specified, appropriate class defaults will be applied automatically.
        type: str
        choices: [Enabled, Disabled]
      cos:
        description:
          - Class of service received by the traffic tagged with this QoS.
          - Valid range is 0-6.
          - FC class defaults to 3.
          - Bronze defaults to 1, Silver to 2, Gold to 4, Platinum to 5.
          - If not specified, appropriate class defaults will be applied automatically.
        type: int
      mtu:
        description:
          - Maximum transmission unit (MTU) is the largest size packet or frame, that can be sent in a packet- or frame-based network such as the Internet.
          - Valid range is 1500-9216.
          - Normal (default) maps to 1500, FC maps to 2240
          - FC class defaults to 2240, others default to 1500.
          - If not specified, appropriate class defaults will be applied automatically.
        type: int
      packet_drop:
        description:
          - If enabled, this QoS class will allow packet drops within an acceptable limit.
          - FC class defaults to false (no packet drop).
          - All other classes default to true (packet drop enabled).
          - If not specified, appropriate class defaults will be applied automatically.
        type: bool
      weight:
        description:
          - The weight of the QoS Class controls the distribution of bandwidth between QoS Classes.
          - Valid range is 0-10.
          - Bronze defaults to 7, Silver to 8, Gold to 9, Platinum to 10, Best Effort and FC to 5.
          - If not specified, appropriate class defaults will be applied automatically.
        type: int

author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Configure System QoS Policy with default classes
  cisco.intersight.intersight_system_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: DevNet
    name: lab-system-qos
    description: System QoS policy for lab use
    tags:
      - Key: Site
        Value: RCDN
    classes:
      - name: Bronze
        admin_state: Enabled
        weight: 7
      - name: Silver
        admin_state: Enabled
        weight: 8
      - name: Gold
        admin_state: Enabled
        weight: 9
      - name: Platinum
        admin_state: Enabled
        weight: 10
      - name: 'Best Effort'
        weight: 5
      - name: FC
        weight: 5

- name: Configure System QoS Policy with custom settings
  cisco.intersight.intersight_system_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: DevNet
    name: custom-system-qos
    description: Custom System QoS policy
    classes:
      - name: Bronze
        admin_state: Enabled
        cos: 1
        mtu: 1500
        packet_drop: true
        weight: 7
      - name: Gold
        admin_state: Enabled
        cos: 4
        mtu: 1500
        packet_drop: true
        weight: 9
      - name: 'Best Effort'
        mtu: 1500
        weight: 6
      - name: FC
        weight: 8

- name: Delete System QoS Policy
  cisco.intersight.intersight_system_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: DevNet
    name: lab-system-qos
    state: absent
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "Name": "lab-system-qos"
    "Description": "System QoS policy for lab use"
    "Organization": {
        "ObjectType": "organization.Organization",
        "Moid": "675450ee69726530014753e2"
    }
    "Classes": [
        {
             "Name": "Bronze",
             "AdminState": "Enabled",
             "Cos": 1,
             "Mtu": 1500,
             "PacketDrop": true,
             "Weight": 7,
             "MulticastOptimize": false
        }
    ]
'''

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec

ALLOWED_CLASSES = ['Bronze', 'Silver', 'Gold', 'Platinum', 'Best Effort', 'FC']


def get_qos_class_defaults(class_name):
    """
    Get default values for QoS classes based on the class name.
    """
    defaults = {
        'Bronze': {
            'admin_state': 'Disabled',
            'cos': 1,
            'mtu': 1500,
            'packet_drop': True,
            'weight': 7
        },
        'Silver': {
            'admin_state': 'Disabled',
            'cos': 2,
            'mtu': 1500,
            'packet_drop': True,
            'weight': 8
        },
        'Gold': {
            'admin_state': 'Disabled',
            'cos': 4,
            'mtu': 1500,
            'packet_drop': True,
            'weight': 9
        },
        'Platinum': {
            'admin_state': 'Disabled',
            'cos': 5,
            'mtu': 1500,
            'packet_drop': True,
            'weight': 10
        },
        'Best Effort': {
            'admin_state': 'Enabled',
            'cos': 255,
            'mtu': 1500,
            'packet_drop': True,
            'weight': 5
        },
        'FC': {
            'admin_state': 'Enabled',
            'cos': 3,
            'mtu': 2240,
            'packet_drop': False,
            'weight': 5
        }
    }
    return defaults.get(class_name, {})


def validate_qos_class(qos_class, module):
    """
    Validate QoS class configuration and apply restrictions.
    """
    class_name = qos_class.get('name')
    if not class_name:
        module.fail_json(msg="QoS class name is required")

    # Apply restrictions for FC class
    if class_name == 'FC':
        if qos_class.get('admin_state') and qos_class['admin_state'] != 'Enabled':
            module.fail_json(msg="FC class must always be enabled")
        if qos_class.get('cos') and qos_class['cos'] != 3:
            module.fail_json(msg="FC class CoS must be 3")
        if qos_class.get('packet_drop') and qos_class['packet_drop']:
            module.fail_json(msg="FC class packet drop must be false")
        if qos_class.get('mtu') and qos_class['mtu'] != 2240:
            module.fail_json(msg="FC class MTU must be 2240")

    # Apply restrictions for Best Effort class
    if class_name == 'Best Effort':
        if qos_class.get('admin_state') and qos_class['admin_state'] != 'Enabled':
            module.fail_json(msg="Best Effort class must always be enabled")
        if qos_class.get('packet_drop') and not qos_class['packet_drop']:
            module.fail_json(msg="Best Effort class packet drop must be true")
        if qos_class.get('cos'):
            module.fail_json(msg="Best Effort class CoS cannot be modified by user")

    # Validate CoS values (Best Effort is handled above)
    if qos_class.get('cos') and class_name != 'Best Effort':
        cos_value = qos_class['cos']
        if cos_value < 0 or cos_value > 6:
            module.fail_json(msg="CoS value must be between 0-6")

    # Validate weight values
    if qos_class.get('weight'):
        weight_value = qos_class['weight']
        if weight_value < 0 or weight_value > 10:
            module.fail_json(msg="Weight value must be between 0-10")

    # Validate MTU values
    if qos_class.get('mtu'):
        mtu_value = qos_class['mtu']
        if mtu_value < 1500 or mtu_value > 9216:
            module.fail_json(msg="MTU value must be between 1500-9216")


def format_qos_classes(classes, module):
    """
    Format QoS classes for API submission, applying defaults and validations.
    """
    if not classes:
        # Return default configuration with all 6 classes
        default_classes = []
        for class_name in ALLOWED_CLASSES:
            defaults = get_qos_class_defaults(class_name)
            default_classes.append({
                'Name': class_name,
                'AdminState': defaults['admin_state'],
                'Cos': defaults['cos'],
                'Mtu': defaults['mtu'],
                'PacketDrop': defaults['packet_drop'],
                'Weight': defaults['weight'],
                'MulticastOptimize': False  # Always false, not user-configurable
            })
        return default_classes

    formatted_classes = []
    provided_class_names = [cls.get('name') for cls in classes]

    # Process provided classes
    for qos_class in classes:
        validate_qos_class(qos_class, module)
        class_name = qos_class['name']
        defaults = get_qos_class_defaults(class_name)

        formatted_class = {
            'Name': class_name,
            'AdminState': qos_class.get('admin_state') if qos_class.get('admin_state') else defaults['admin_state'],
            'Cos': qos_class.get('cos') if qos_class.get('cos') else defaults['cos'],
            'Mtu': qos_class.get('mtu') if qos_class.get('mtu') else defaults['mtu'],
            'PacketDrop': qos_class.get('packet_drop') if qos_class.get('packet_drop') else defaults['packet_drop'],
            'Weight': qos_class.get('weight') if qos_class.get('weight') else defaults['weight'],
            'MulticastOptimize': False  # Always false, not user-configurable
        }
        formatted_classes.append(formatted_class)

    # Add missing classes with defaults
    for class_name in ALLOWED_CLASSES:
        if class_name not in provided_class_names:
            defaults = get_qos_class_defaults(class_name)
            formatted_classes.append({
                'Name': class_name,
                'AdminState': defaults['admin_state'],
                'Cos': defaults['cos'],
                'Mtu': defaults['mtu'],
                'PacketDrop': defaults['packet_drop'],
                'Weight': defaults['weight'],
                'MulticastOptimize': False  # Always false, not user-configurable
            })

    return formatted_classes


def main():
    class_spec = dict(
        name=dict(type='str', required=True, choices=ALLOWED_CLASSES),
        admin_state=dict(type='str', choices=['Enabled', 'Disabled']),
        cos=dict(type='int'),
        mtu=dict(type='int'),
        packet_drop=dict(type='bool'),
        weight=dict(type='int')
    )

    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        organization=dict(type='str', default='default'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr']),
        tags=dict(type='list', elements='dict'),
        classes=dict(type='list', elements='dict', options=class_spec)
    )

    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )

    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''

    # Resource path used to configure policy
    resource_path = '/fabric/SystemQosPolicies'
    # Define API body used in compares or create
    intersight.api_body = {
        'Organization': {
            'Name': intersight.module.params['organization']
        },
        'Name': intersight.module.params['name']
    }

    if module.params['state'] == 'present':
        # Format QoS classes with validation and defaults
        formatted_classes = format_qos_classes(module.params.get('classes'), module)
        intersight.api_body['Classes'] = formatted_classes
        intersight.set_tags_and_description()

    intersight.configure_policy_or_profile(resource_path=resource_path)

    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "certified"}

DOCUMENTATION = r"""
---
module: aci_contract_subject_to_filter
short_description: Bind Contract Subjects to Filters (vz:RsSubjFiltAtt)
description:
- Bind Contract Subjects to Filters on Cisco ACI fabrics.
options:
  contract:
    description:
    - The name of the contract.
    type: str
    aliases: [ contract_name ]
  contract_type:
    description:
    - The type of contract, either standard or Out of Band (oob).
    type: str
    choices: [ standard, oob ]
    default: standard
  filter:
    description:
    - The name of the Filter to bind to the Subject.
    type: str
    aliases: [ filter_name ]
  direction:
    description:
    - The direction of traffic matching for the filter.
    type: str
    default: both
    choices: [ both, consumer_to_provider, provider_to_consumer ]
  action:
    description:
    - The action required when the condition is met.
    - The APIC defaults to C(permit) when unset during creation.
    type: str
    choices: [ deny, permit ]
  priority_override:
    description:
    - Overrides the filter priority for the a single applied filter.
    type: str
    choices: [ default, level1, level2, level3 ]
  directives:
    description:
    - Determines if the binding should be set to log.
    - The APIC defaults to C(none) when unset during creation.
    type: str
    choices: [ log, no_stats, none ]
    aliases: [ log, directive]
  subject:
    description:
    - The name of the Contract Subject.
    type: str
    aliases: [ contract_subject, subject_name ]
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
  tenant:
    description:
    - The name of the tenant.
    type: str
    aliases: [ tenant_name ]
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant), C(contract), C(subject), and C(filter_name) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant), M(cisco.aci.aci_contract), M(cisco.aci.aci_contract_subject), and M(cisco.aci.aci_filter) modules can be used for these.
seealso:
- module: cisco.aci.aci_contract_subject
- module: cisco.aci.aci_filter
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(vz:RsSubjFiltAtt).
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Jacob McGill (@jmcgill298)
"""

EXAMPLES = r"""
- name: Add a new contract subject to filer binding
  cisco.aci.aci_contract_subject_to_filter:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web_to_db
    subject: test
    filter: '{{ filter }}'
    log: '{{ log }}'
    state: present
  delegate_to: localhost

- name: Remove an existing contract subject to filter binding
  cisco.aci.aci_contract_subject_to_filter:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web_to_db
    subject: test
    filter: '{{ filter }}'
    log: '{{ log }}'
    state: present
  delegate_to: localhost

- name: Query a specific contract subject to filter binding
  cisco.aci.aci_contract_subject_to_filter:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web_to_db
    subject: test
    filter: '{{ filter }}'
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all contract subject to filter bindings
  cisco.aci.aci_contract_subject_to_filter:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: production
    contract: web_to_db
    subject: test
    state: query
  delegate_to: localhost
  register: query_result
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec
from ansible_collections.cisco.aci.plugins.module_utils.constants import CONTRACT_CLASS_MAPPING


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"]),  # Not required for querying all objects
        contract=dict(type="str", aliases=["contract_name"]),  # Not required for querying all objects
        contract_type=dict(type="str", choices=["standard", "oob"], default="standard"),
        filter=dict(type="str", aliases=["filter_name"]),  # Not required for querying all objects
        subject=dict(type="str", aliases=["contract_subject", "subject_name"]),  # Not required for querying all objects
        # default both because of back-worth compatibility and for determining which config to push
        direction=dict(type="str", default="both", choices=["both", "consumer_to_provider", "provider_to_consumer"]),
        action=dict(type="str", choices=["deny", "permit"]),
        # named directives instead of log/directive for readability of code, aliases and input "none are kept for back-worth compatibility
        directives=dict(type="str", choices=["log", "no_stats", "none"], aliases=["log", "directive"]),
        priority_override=dict(type="str", choices=["default", "level1", "level2", "level3"]),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["contract", "filter", "subject", "tenant"]],
            ["state", "present", ["contract", "filter", "subject", "tenant"]],
        ],
    )

    contract = module.params.get("contract")
    contract_type = module.params.get("contract_type")
    filter_name = module.params.get("filter")
    # "none" is kept because of back-worth compatibility, could be deleted and keep only None
    directives = "" if (module.params.get("directives") is None or module.params.get("directives") == "none") else module.params.get("directives")
    subject = module.params.get("subject")
    direction = module.params.get("direction")
    action = module.params.get("action")
    priority_override = module.params.get("priority_override")
    tenant = module.params.get("tenant")
    state = module.params.get("state")

    base_subject_dict = dict(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class=CONTRACT_CLASS_MAPPING[contract_type]["class"],
            aci_rn=CONTRACT_CLASS_MAPPING[contract_type]["rn"].format(contract),
            module_object=contract,
            target_filter={"name": contract},
        ),
        subclass_2=dict(
            aci_class="vzSubj",
            aci_rn="subj-{0}".format(subject),
            module_object=subject,
            target_filter={"name": subject},
        ),
    )

    aci = ACIModule(module)

    # start logic to be consistent with GUI to only allow both direction or a one-way connection
    aci.construct_url(
        root_class=base_subject_dict.get("root_class"),
        subclass_1=base_subject_dict.get("subclass_1"),
        subclass_2=base_subject_dict.get("subclass_2"),
        child_classes=["vzInTerm", "vzOutTerm"],
    )
    aci.get_existing()
    direction_options = ["both"]
    if aci.existing:
        direction_options = ["consumer_to_provider", "provider_to_consumer"] if "children" in aci.existing[0]["vzSubj"] else ["both"]

    if state != "query" and direction not in direction_options:
        module.fail_json(msg="Direction is not allowed, valid option is {0}.".format(" or ".join(direction_options)))
    # end logic to be consistent with GUI to only allow both direction or a one-way connection

    if direction == "both":
        filter_class = "vzRsSubjFiltAtt"
        # dict unpacking with **base_subject_dict raises syntax error in python2.7 thus dict lookup
        aci.construct_url(
            root_class=base_subject_dict.get("root_class"),
            subclass_1=base_subject_dict.get("subclass_1"),
            subclass_2=base_subject_dict.get("subclass_2"),
            subclass_3=dict(
                aci_class=filter_class,
                aci_rn="rssubjFiltAtt-{0}".format(filter_name),
                module_object=filter_name,
                target_filter=dict(tnVzFilterName=filter_name),
            ),
        )
    else:
        term_class, term = ("vzInTerm", "intmnl") if direction == "consumer_to_provider" else ("vzOutTerm", "outtmnl")
        filter_class = "vzRsFiltAtt"
        # dict unpacking with **base_subject_dict raises syntax error in python2.7 thus dict lookup
        aci.construct_url(
            root_class=base_subject_dict.get("root_class"),
            subclass_1=base_subject_dict.get("subclass_1"),
            subclass_2=base_subject_dict.get("subclass_2"),
            subclass_3=dict(aci_class=term_class, aci_rn=term),
            child_classes=[filter_class],
        )

    aci.get_existing()

    if state == "present":
        config = dict(tnVzFilterName=filter_name, directives=directives, action=action, priorityOverride=priority_override)
        if direction == "both":
            aci.payload(aci_class=filter_class, class_config=config)
            aci.get_diff(aci_class=filter_class)
        else:
            child_config = [dict(vzRsFiltAtt=dict(attributes=config))]
            aci.payload(aci_class=term_class, class_config=dict(), child_configs=child_config)
            aci.get_diff(aci_class=term_class)
        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    if direction == "both":
        aci.exit_json()
    else:
        # filter the output of current/previous to tnVzFilterName only since existing consist full vzInTerm/vzOutTerm
        def filter_result(input_list, name):
            return [
                {key: filter_entry}
                for entry in input_list
                if "children" in entry[term_class]
                for children in entry[term_class]["children"]
                for key, filter_entry in children.items()
                if filter_entry["attributes"]["tnVzFilterName"] == name
            ]

        # pass function to
        filter_existing = (filter_result, filter_name)
        aci.exit_json(filter_existing)


if __name__ == "__main__":
    main()



# Copyright (c) 2016, Hiroaki Nakamura <hnakamur@gmail.com>
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations

import http.client as http_client
import json
import os
import socket
import ssl
import typing as t
from urllib.parse import urlparse

from ansible.module_utils.urls import generic_urlparse

# httplib/http.client connection using unix domain socket
HTTPConnection = http_client.HTTPConnection
HTTPSConnection = http_client.HTTPSConnection


class UnixHTTPConnection(HTTPConnection):
    def __init__(self, path: str) -> None:
        HTTPConnection.__init__(self, "localhost")
        self.path = path

    def connect(self):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.connect(self.path)
        self.sock = sock


class LXDClientException(Exception):
    def __init__(self, msg: str, **kwargs) -> None:
        self.msg = msg
        self.kwargs = kwargs


class LXDClient:
    def __init__(
        self,
        url: str,
        key_file: str | None = None,
        cert_file: str | None = None,
        debug: bool = False,
        server_cert_file: str | None = None,
        server_check_hostname: bool = True,
    ) -> None:
        """LXD Client.

        :param url: The URL of the LXD server. (e.g. unix:/var/lib/lxd/unix.socket or https://127.0.0.1)
        :param key_file: The path of the client certificate key file.
        :param cert_file: The path of the client certificate file.
        :param debug: The debug flag. The request and response are stored in logs when debug is true.
        :param server_cert_file: The path of the server certificate file.
        :param server_check_hostname: Whether to check the server's hostname as part of TLS verification.
        """
        self.url = url
        self.debug = debug
        self.logs: list[dict[str, t.Any]] = []
        self.connection: UnixHTTPConnection | HTTPSConnection
        if url.startswith("https:"):
            self.cert_file = cert_file
            self.key_file = key_file
            parts = generic_urlparse(urlparse(self.url))
            ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            if server_cert_file:
                # Check that the received cert is signed by the provided server_cert_file
                ctx.load_verify_locations(cafile=server_cert_file)
            ctx.check_hostname = server_check_hostname
            ctx.load_cert_chain(cert_file, keyfile=key_file)  # type: ignore # TODO!
            self.connection = HTTPSConnection(parts.get("netloc"), context=ctx)
        elif url.startswith("unix:"):
            unix_socket_path = url[len("unix:") :]
            self.connection = UnixHTTPConnection(unix_socket_path)
        else:
            raise LXDClientException("URL scheme must be unix: or https:")

    def do(self, method: str, url: str, body_json=None, ok_error_codes=None, timeout=None, wait_for_container=None):
        resp_json = self._send_request(method, url, body_json=body_json, ok_error_codes=ok_error_codes, timeout=timeout)
        if resp_json["type"] == "async":
            url = f"{resp_json['operation']}/wait"
            resp_json = self._send_request("GET", url)
            if wait_for_container:
                while resp_json["metadata"]["status"] == "Running":
                    resp_json = self._send_request("GET", url)
            if resp_json["metadata"]["status"] != "Success":
                self._raise_err_from_json(resp_json)
        return resp_json

    def authenticate(self, trust_password):
        body_json = {"type": "client", "password": trust_password}
        return self._send_request("POST", "/1.0/certificates", body_json=body_json)

    def _send_request(self, method: str, url: str, body_json=None, ok_error_codes=None, timeout=None):
        try:
            body = json.dumps(body_json)
            self.connection.request(method, url, body=body)
            resp = self.connection.getresponse()
            resp_data = resp.read()
            resp_json = json.loads(resp_data)
            self.logs.append(
                {
                    "type": "sent request",
                    "request": {"method": method, "url": url, "json": body_json, "timeout": timeout},
                    "response": {"json": resp_json},
                }
            )
            resp_type = resp_json.get("type", None)
            if resp_type == "error":
                if ok_error_codes is not None and resp_json["error_code"] in ok_error_codes:
                    return resp_json
                if resp_json["error"] == "Certificate already in trust store":
                    return resp_json
                self._raise_err_from_json(resp_json)
            return resp_json
        except OSError as e:
            raise LXDClientException("cannot connect to the LXD server", err=e) from e

    def _raise_err_from_json(self, resp_json):
        err_params = {}
        if self.debug:
            err_params["logs"] = self.logs
        raise LXDClientException(self._get_err_from_resp_json(resp_json), **err_params)

    @staticmethod
    def _get_err_from_resp_json(resp_json):
        err = None
        metadata = resp_json.get("metadata", None)
        if metadata is not None:
            err = metadata.get("err", None)
        if err is None:
            err = resp_json.get("error", None)
        return err


def default_key_file() -> str:
    return os.path.expanduser("~/.config/lxc/client.key")


def default_cert_file() -> str:
    return os.path.expanduser("~/.config/lxc/client.crt")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# (c) 2017, Ansible by Red Hat, Inc
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type


DOCUMENTATION = """
module: iosxr_netconf
author: Kedar Kekan (@kedarX)
short_description: Configures NetConf sub-system service on Cisco IOS-XR devices
description:
- This module provides an abstraction that enables and configures the netconf system
  service running on Cisco IOS-XR Software. This module can be used to easily enable
  the Netconf API. Netconf provides a programmatic interface for working with configuration
  and state resources as defined in RFC 6242.
version_added: 1.0.0
extends_documentation_fragment:
- cisco.iosxr.iosxr
options:
  netconf_port:
    description:
    - This argument specifies the port the netconf service should listen on for SSH
      connections.  The default port as defined in RFC 6242 is 830.
    required: false
    type: int
    default: 830
    aliases:
    - listens_on
  netconf_vrf:
    description:
    - netconf vrf name
    required: false
    type: str
    default: default
    aliases:
    - vrf
  state:
    description:
    - Specifies the state of the C(iosxr_netconf) resource on the remote device.  If
      the I(state) argument is set to I(present) the netconf service will be configured.  If
      the I(state) argument is set to I(absent) the netconf service will be removed
      from the configuration.
    type: str
    required: false
    default: present
    choices:
    - present
    - absent
notes:
- This module works with connection C(network_cli). See L(the IOS-XR Platform Options,../network/user_guide/platform_iosxr.html).
"""

EXAMPLES = """
- name: enable netconf service on port 830
  cisco.iosxr.iosxr_netconf:
    listens_on: 830
    state: present

- name: disable netconf service
  cisco.iosxr.iosxr_netconf:
    state: absent
"""

RETURN = """
commands:
  description: Returns the command sent to the remote device
  returned: when changed is True
  type: str
  sample: 'ssh server netconf port 830'
"""
import re

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.iosxr.plugins.module_utils.network.iosxr.iosxr import (
    get_config,
    load_config,
)


USE_PERSISTENT_CONNECTION = True


def map_obj_to_commands(updates):
    want, have = updates
    commands = list()

    if want["state"] == "absent":
        if have["state"] == "present":
            commands.append("no netconf-yang agent ssh")

        if "netconf_port" in have:
            commands.append("no ssh server netconf port %s" % have["netconf_port"])

        if have["netconf_vrf"]:
            for vrf in have["netconf_vrf"]:
                commands.append("no ssh server netconf vrf %s" % vrf)
    else:
        if have["state"] == "absent":
            commands.append("netconf-yang agent ssh")

        if want["netconf_port"] is not None and (want["netconf_port"] != have.get("netconf_port")):
            commands.append("ssh server netconf port %s" % want["netconf_port"])
        if want["netconf_vrf"] is not None and (want["netconf_vrf"] not in have["netconf_vrf"]):
            commands.append("ssh server netconf vrf %s" % want["netconf_vrf"])

    return commands


def parse_vrf(config):
    match = re.search(r"vrf (\w+)", config)
    if match:
        return match.group(1)


def parse_port(config):
    match = re.search(r"port (\d+)", config)
    if match:
        return int(match.group(1))


def map_config_to_obj(module):
    obj = {"state": "absent"}

    netconf_config = get_config(module, config_filter="netconf-yang agent")

    ssh_config = get_config(module, config_filter="ssh server")
    ssh_config = [
        config_line
        for config_line in (line.strip() for line in ssh_config.splitlines())
        if config_line
    ]
    obj["netconf_vrf"] = []
    for config in ssh_config:
        if "netconf port" in config:
            obj.update({"netconf_port": parse_port(config)})
        if "netconf vrf" in config:
            obj["netconf_vrf"].append(parse_vrf(config))
    if "ssh" in netconf_config and ("netconf_port" in obj or obj["netconf_vrf"]):
        obj.update({"state": "present"})

    if "ssh" in netconf_config and "netconf_port" not in obj:
        obj.update({"netconf_port": 830})

    return obj


def validate_netconf_port(value, module):
    if not 1 <= value <= 65535:
        module.fail_json(msg="netconf_port must be between 1 and 65535")


def map_params_to_obj(module):
    obj = {
        "netconf_port": module.params["netconf_port"],
        "netconf_vrf": module.params["netconf_vrf"],
        "state": module.params["state"],
    }

    for key, value in obj.items():
        # validate the param value (if validator func exists)
        validator = globals().get("validate_%s" % key)
        if callable(validator):
            validator(value, module)

    return obj


def main():
    """main entry point for module execution"""
    argument_spec = dict(
        netconf_port=dict(type="int", default=830, aliases=["listens_on"]),
        netconf_vrf=dict(aliases=["vrf"], default="default"),
        state=dict(default="present", choices=["present", "absent"]),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    warnings = list()

    result = {"changed": False, "warnings": warnings}

    want = map_params_to_obj(module)
    have = map_config_to_obj(module)
    commands = map_obj_to_commands((want, have))
    result["commands"] = commands

    if commands:
        commit = not module.check_mode
        diff = load_config(module, commands, commit=commit)
        if diff:
            result["diff"] = dict(prepared=diff)
        result["changed"] = True

    module.exit_json(**result)


if __name__ == "__main__":
    main()
